"""
分箱逻辑：
    1、类别型特征
      1）类别数在5个以下，可以直接根据类别来分箱 (binning_categorical)
      2）类别数在5个以上，建议做降基处理，再根据降基后的类别做分箱(binning_categorical)
    2、数值型特征
      1）无顺序型数值特征：
        若特征value的非重复计数在5个以下，可以直接根据非重复计数值来分箱(binning_categorical)
        若特征value的非重复计数在5个以上，建议根据业务解释或者数据分布做自定义分箱(binning_custom)
      2）连续型数值特征：
        可以用卡方分箱或自定义分箱。(binning_continuous, binning_custom)
        PS:一些特征用卡方分可能会报错，建议这些特征改为手动自定义分箱(binning_custom))
    3、其他
      1）缺失率在5%以下，可以先对缺失做填充处理再分箱
      2）缺失率在5%以上，建议将缺失当作一个类别来分箱
      3) 建议将稀疏值（一般为0）单独分为一箱
"""
from abc import ABCMeta, abstractmethod
import numpy as np
import pandas as pd
from . import utils


class Discretizer(object, metaclass=ABCMeta):
    def __init__(self, continuous_flag,
                 spe_value_list, spe_bin_indicator):
        self._spe_value_list = spe_value_list
        self._spe_bin_indicator = spe_bin_indicator
        self._continuous_flag = continuous_flag
        self._bin_df = None
        self._change_bin_num = False

    @property
    def model(self):
        return self._bin_df

    @abstractmethod
    def _fit_core(self, df, col, target):
        pass

    def fit(self, df, col, target=None):
        if self._continuous_flag is None:
            if df[col].dtypes == 'object':
                self._continuous_flag = False
            else:
                self._continuous_flag = True
        if len(self._spe_value_list) >= 1:
            df1 = df.loc[df[col].isin(self._spe_value_list)]
            df2 = df.loc[~df[col].isin(self._spe_value_list)]
            if df1.shape[0] != 0 and self._spe_bin_indicator:
                self._change_bin_num = True
            else:
                self._change_bin_num = False
        else:
            df2 = df.copy()

        # 对于无监督学习，dataframe[bin_range, bin_total]
        # 对于有监督学习，dataframe[bin_range, bin_total,
        #                        bin_good, bin_bad, bin_bad_rate]
        if df2.shape[0] == 0:
            print('no effective data, failure')
            bin_df = None
        else:
            bin_df = self._fit_core(df2, col, target)
        if bin_df is not None:
            if len(self._spe_value_list) >= 1 \
                    and self._spe_bin_indicator is True:
                if target is None:
                    regroup_single = pd.DataFrame()
                    regroup_single['bin_total'] = df1.groupby(col).size()
                    # dataframe [col, bin_total]
                    regroup_single.reset_index(inplace=True)
                else:
                    # dataframe [col, bin_total, bin_good, bin_bad, bin_bad_rate]
                    regroup_single = utils.bin_bad_rate(df1, col, target)
                regroup_single.rename(columns={col: 'bin_range'}, inplace=True)
                bin_df = utils.concat_of_spec_bins(regroup_single, bin_df,
                                                   'bin_range', self._continuous_flag)

            if target is not None:
                bin_df = utils.compute_woe_iv(bin_df, 'bin_good', 'bin_bad', 'bin_total')

            # 在最后两列生成feature_name和bin_no
            bin_df['feature_name'] = col
            if self._change_bin_num:
                bin_df['bin_no'] = np.array(
                    self._spe_value_list + list(range(
                        bin_df.shape[0] - len(self._spe_value_list)))
                ).astype('object')
            else:
                bin_df['bin_no'] = np.arange(bin_df.shape[0], dtype=float)
            # 将最后两列调整到第前面
            bin_df = pd.concat(
                [bin_df[['feature_name', 'bin_no']], bin_df.iloc[:, :-2]], axis=1)

        # 对于无监督学习，dataframe[feature_name, bin_range, bin_total, bin_no]
        # 对于有监督学习，dataframe[feature_name, bin_range, bin_total, bin_good,
        #                        bin_bad, bin_bad_rate, bin_no, bin_woe, iv]
        # index一定是从0开始逐渐增大
        self._bin_df = bin_df

    def transform(self, df, from_col, to_col, test_indicator=False,
                  target=None, fill_na=None, failure_value=0):
        assert self._bin_df is not None
        if fill_na is None and len(self._spe_value_list) != 0:
            fill_na = self._spe_value_list[0]

        if test_indicator:
            mapped_col, bin_df2 = \
                utils.map_to_bin(df, from_col, self._bin_df, to_col,
                                 test_indicator, target, fill_na, failure_value)
            return mapped_col, bin_df2
        else:
            mapped_col = utils.map_to_bin(df, from_col, self._bin_df, to_col,
                                          test_indicator, target, fill_na, failure_value)
            return mapped_col

    def combine_bin_to_next(self, bin_df, index, inplace=False):
        """
        :param bin_df: 对于无监督学习，dataframe[feature_name, bin_range, bin_total, bin_no]
                       对于有监督学习，dataframe[feature_name, bin_range, bin_total, bin_good,
                                              bin_bad, bin_bad_rate, bin_no, bin_woe, iv]
                       bin_no和indices永远是从0开始递增
        :param index: int, rows of index and index+1 of bin_df will be combined
        :param inplace: bool, whether to update the mapping model
        """
        columns = bin_df.columns.tolist()
        # make sure bin_df is [feature_name, bin_range, bin_total] or
        # [feature_name, bin_range, bin_total, bin_good, bin_bad, bin_bad_rate]
        for col_name in ['bin_no', 'bin_woe', 'iv']:
            if col_name in columns:
                bin_df = bin_df.drop(col_name, axis=1)
        bin_df = utils.combine_bin_df(bin_df, 'bin_range', index, self._continuous_flag)
        if self._change_bin_num:
            bin_df['bin_no'] = np.array(
                self._spe_value_list + list(range(
                    bin_df.shape[0] - len(self._spe_value_list)))
            ).astype('object')
        else:
            bin_df['bin_no'] = np.arange(bin_df.shape[0], dtype=float)
        if 'bin_woe' in columns:
            bin_df = utils.compute_woe_iv(bin_df, 'bin_good', 'bin_bad', 'bin_total')
        if inplace:
            self._bin_df = bin_df
        return bin_df


class SimpleEncoder(Discretizer):
    def __init__(self, spe_value_list=[], spe_bin_indicator=False, continuous_flag=False,
                 min_binpct=0.05, monoticity=False, auto_merge=True):
        super(SimpleEncoder, self).__init__(False,
                                            spe_value_list,
                                            spe_bin_indicator)
        self._auto_merge = auto_merge
        self._min_binpct = min_binpct
        self._monoticity = monoticity
        self._continuous_flag = continuous_flag

    def _fit_core(self, df, col, target=None):
        if target is None:
            bin_df = pd.DataFrame()
            bin_df['bin_total'] = df.groupby(col).size()
            bin_df.reset_index(inplace=True)
            # dataframe [col, bin_total]
        else:
            # dataframe [col, bin_total, bin_good, bin_bad, bin_bad_rate]
            # index一定是从0开始逐渐编号
            bin_df = utils.bin_bad_rate(df, col, target)
            bin_df[col] = bin_df[col].apply(lambda x: [x])
            if self._continuous_flag:
                group_interval = bin_df[col].tolist()
                if bin_df.shape[0] == 1:
                    cut_off_points = [max(group_interval[0])]
                    bin_df[col] = [pd.Interval(left=float('-inf'),
                                               right=float(cut_off_points[0]))]
                else:
                    # 取每个区间的最大值, cutoffpoints不含起止点
                    cut_off_points = [max(i) if isinstance(i, list) else i
                                      for i in group_interval[:-1]]
                    bin_df[col] = utils.cutoffpoints_to_interval(cut_off_points)
            else:
                # 对于类别型变量需要按坏账率排序，合并相似坏账率的组
                bin_df.sort_values(by='bin_bad_rate', inplace=True)
                bin_df.reset_index(inplace=True, drop=True)
            if self._auto_merge:
                bin_df = utils.check_bin(bin_df, col, self._continuous_flag,
                                         self._min_binpct, self._monoticity)
                if bin_df is None:
                    return None
        bin_df.rename(columns={col: 'bin_range'}, inplace=True)
        return bin_df


class CommonNumericalDiscretizer(Discretizer):
    def __init__(self, spe_value_list=[], spe_bin_indicator=False,
                 cut=5, quantile=True, min_binpct=0, monoticity=False,
                 auto_merge=True):
        super(CommonNumericalDiscretizer, self).__init__(True,
                                                         spe_value_list,
                                                         spe_bin_indicator)
        self._auto_merge = auto_merge
        self._min_binpct = min_binpct
        self._monoticity = monoticity
        self._quantile = quantile
        if len(spe_value_list) >= 1 and spe_bin_indicator is True:
            self._cut = cut - len(spe_value_list)
        else:
            self._cut = cut

    def _fit_core(self, df, col, target=None):
        if self._quantile:
            bucket = pd.qcut(df[col], self._cut, duplicates='drop')
        else:
            bucket = pd.cut(df[col], self._cut)
        if target is None:
            bin_df = pd.DataFrame()
            bin_df['bin_total'] = df.groupby(bucket.rename('bin_range')).size()
            # dataframe [bin_range, bin_total]
            # index一定是从0开始逐渐编号
            bin_df.reset_index(inplace=True)
        else:
            # dataframe [bin_range, bin_total, bin_good, bin_bad, bin_bad_rate]
            # index一定是从0开始逐渐编号
            bin_df = utils.bin_bad_rate(df, bucket.rename('bin_range'), target)
            if self._auto_merge:
                bin_df = utils.check_bin(bin_df, 'bin_range', True,
                                         self._min_binpct, self._monoticity)
                if bin_df is None:
                    return None
        bin_df = utils.map_to_inf(bin_df, 'bin_range')
        return bin_df


class CommonCategoricalDiscretizer(Discretizer):
    def __init__(self, spe_value_list=[], spe_bin_indicator=False, monoticity=False,
                 cut=5, quantile=True, min_binpct=0, auto_merge=True):
        super(CommonCategoricalDiscretizer, self).__init__(False,
                                                           spe_value_list,
                                                           spe_bin_indicator)
        self._auto_merge = auto_merge
        self._min_binpct = min_binpct
        self._quantile = quantile
        self._monoticity = monoticity
        if len(spe_value_list) >= 1 and spe_bin_indicator is True:
            self._cut = cut - len(spe_value_list)
        else:
            self._cut = cut

    def _fit_core(self, df, col, target=None):
        # 类别型变量往往取值极其不平衡，而且样本数目远远大于类别数目
        # 如果在原始数据上直接取等频分位点的话，会出现大量重复的分割点，
        # 这样也不能等频划分为你想要的区间数目。
        # 因此，如果对类别取值去重，然后取分位点。这样处理的后果是：
        # 1. 等宽划分，完全不受去重的影响，因为等宽分位点是按取值的大小进行的
        # 2. 等频分位点是按序列的index取的，因此去重后的index是不一样的，
        #    划分的区间映射回原始数据集，各个区间的样本个数也是不同的
        #    最后的结果和等宽划分差别并不大
        if target is None:
            map_dict = df.groupby(col)[col].count().to_dict()
            df.loc[:, 'proxy'] = df[col].map(map_dict)
        else:
            regroup = utils.bin_bad_rate(df, col, target)
            map_dict = regroup.set_index(col)['bin_bad_rate'].to_dict()
            df.loc[:, 'proxy'] = df[col].map(map_dict)
        if not self._quantile:
            bucket = pd.cut(df['proxy'], self._cut)
        else:
            bucket = pd.qcut(df['proxy'], self._cut, duplicates='drop')
        # iterable of arrays
        indices_list = df.groupby(bucket).indices.values()
        bin_df = pd.DataFrame()
        # 在未排序的dataframe中，按照indices来取数,
        # 由于indices是array，因此得到的一定是series,
        # series.unique()得到的一定是array且tolist()以后一定是list
        # dataframe [bin_range, bin_total]
        col_values = [df.loc[df.index[indices], col].unique().tolist() for indices in indices_list
                      if indices.shape[0] != 0]
        bin_df['bin_range'] = col_values
        col_total = [df.loc[df.index[indices], col].size for indices in indices_list
                     if indices.shape[0] != 0]
        bin_df['bin_total'] = col_total
        if target is not None:
            # dataframe [bin_range, bin_total, bin_good, bin_bad, bin_bad_rate]
            # index一定是从0开始逐渐增大的
            col_bad = [df.loc[df.index[indices], target].sum() for indices in indices_list
                       if indices.shape[0] != 0]
            bin_df['bin_bad'] = col_bad
            bin_df['bin_good'] = bin_df['bin_total'] - bin_df['bin_bad']
            bin_df['bin_bad_rate'] = bin_df['bin_bad'] / bin_df['bin_total']
            if self._auto_merge:
                bin_df = utils.check_bin(bin_df, 'bin_range', False,
                                         self._min_binpct, self._monoticity)
                if bin_df is None:
                    return None
        return bin_df


class Chi2Discretizer(Discretizer):
    def __init__(self, spe_value_list=[], spe_bin_indicator=False,
                 continuous_flag=True, precut=100, prequantile=True,
                 max_bin_num=5, min_binpct=0.0, monoticity=True):
        super(Chi2Discretizer, self).__init__(continuous_flag,
                                              spe_value_list,
                                              spe_bin_indicator)
        self._monoticity = monoticity
        self._precutthreshold = precut
        self._prequantile = prequantile
        self._min_binpct = min_binpct
        self._max_bin_num = max_bin_num

    def _fit_core(self, df, col, target):
        if len(self._spe_value_list) >= 1 and self._spe_bin_indicator is True and self._change_bin_num:
            max_bin_num = self._max_bin_num - len(self._spe_value_list)
        else:
            max_bin_num = self._max_bin_num
        if self._continuous_flag:
            if len(df[col].unique()) > self._precutthreshold:
                # 对于连续型变量，如果唯一值数目过大，
                # 则先做一次等宽或等频分箱，把原col映射到盖帽分割点上（注意不是分箱号)
                split_points = utils.unsupervised_split(df, col, self._precutthreshold, self._prequantile)
                df[col] = df[col].map(lambda x: utils.cutoffpoints_to_edge(x, split_points))
            bin_df = utils.bin_bad_rate(df, col, target)
            bin_df[col] = bin_df[col].apply(lambda x: [x])
        else:
            if len(df[col].unique()) > self._precutthreshold:
                preprocessor = CommonCategoricalDiscretizer(spe_value_list=[],
                                                            spe_bin_indicator=False,
                                                            cut=self._precutthreshold,
                                                            quantile=self._prequantile,
                                                            min_binpct=0,
                                                            monoticity=False,
                                                            auto_merge=False)
                preprocessor.fit(df, col, target)
                # index一定是从0开始逐渐增大
                bin_df = preprocessor.model[['bin_range', 'bin_total', 'bin_bad',
                                             'bin_good', 'bin_bad_rate']].rename(
                    columns={'bin_range': col})
            else:
                bin_df = utils.bin_bad_rate(df, col, target)
                bin_df[col] = bin_df[col].apply(lambda x: [x])
                bin_df.sort_values(by='bin_bad_rate', inplace=True)
                # index一定是从0开始逐渐增大
                bin_df.reset_index(inplace=True, drop=True)

        bin_df = utils.chi2_merge(bin_df, col, max_bin_num,
                                  self._min_binpct, self._monoticity)

        if self._continuous_flag:
            group_interval = bin_df[col].tolist()
            if bin_df.shape[0] == 1:
                cut_off_points = [max(group_interval[0])]
                bin_df[col] = [pd.Interval(left=float('-inf'),
                                           right=float(cut_off_points[0]))]
            else:
                # 取每个区间的最大值, cutoffpoints不含起止点
                cut_off_points = [max(i) if isinstance(i, list) else i
                                  for i in group_interval[:-1]]
                bin_df[col] = utils.cutoffpoints_to_interval(cut_off_points)

        # dataframe [bin_range, bin_total, bin_bad, bin_good, bin_bad_rate]
        bin_df.rename(columns={col: 'bin_range'}, inplace=True)
        return bin_df


class Chi2Discretizer2(Discretizer):
    def __init__(self, bc_iv_lower, bc_iv_upper, spe_value_list=[], spe_bin_indicator=False,
                 continuous_flag=True, max_bin_num=5, min_binpct=0.1, monotonicity=True,
                 auto_merge=True):
        assert continuous_flag is True
        assert 0.05 <= min_binpct < 1
        super(Chi2Discretizer2, self).__init__(continuous_flag,
                                               spe_value_list,
                                               spe_bin_indicator)
        self._monotonicity = monotonicity
        self._bc_iv_lower = bc_iv_lower
        self._bc_iv_upper = bc_iv_upper
        self._min_binpct = min_binpct
        self._max_bin_num = max_bin_num
        self._auto_merge = auto_merge

    def _fit_core(self, df, col, target):
        if len(self._spe_value_list) >= 1 and self._spe_bin_indicator is True and self._change_bin_num:
            max_bin_num = self._max_bin_num - len(self._spe_value_list)
        else:
            max_bin_num = self._max_bin_num
        total = df.shape[0]
        bad = df[target].sum()
        good = total - bad
        bc_good_bad = (good, bad)
        bc_gap = total * self._min_binpct
        df['temp'] = col
        group = df.loc[:, ['temp', col, target]].values.tolist()
        bin_df = utils.best_chisq_bin(group, max_bin_num, bc_good_bad, bc_gap, self._monotonicity,
                                      self._bc_iv_lower, self._bc_iv_upper, self._auto_merge)
        return bin_df





from statsmodels.discrete.discrete_model import Logit
from statsmodels.tools.tools import add_constant
from scipy import stats
from sklearn import linear_model
import pandas as pd
import numpy as np
stats.chisqprob = lambda chisq, df: stats.chi2.sf(chisq, df)


def calculate_partial_correlation(input_df, column1, column2):
    """
    Returns linear partial correlation coefficients
    between pairs of variables,
    controlling all the other variables

    Parameters
    ----------
    input_df : pandas.DataFrame, shape (n, p)
        Array with candidate variables. Each column is taken as a variable.

    Returns
    -------
    P :  partial correlation of input_df[:, column1] and input_df[:, column2]
        controlling for all other remaining variables.
    """
    if column1 == column2:
        return 1
    control_variables = input_df.columns.tolist()
    control_variables.remove(column1)
    control_variables.revove(column2)
    data_control_variable = input_df.loc[:, control_variables]
    data_column1 = input_df[column1].values
    data_column2 = input_df[column2].values
    fit1 = linear_model.LinearRegression(fit_intercept=True)
    fit2 = linear_model.LinearRegression(fit_intercept=True)
    fit1.fit(data_control_variable, data_column1)
    fit2.fit(data_control_variable, data_column2)
    residual1 = data_column1 - (
            np.dot(data_control_variable, fit1.coef_) + fit1.intercept_)
    residual2 = data_column2 - (
            np.dot(data_control_variable, fit2.coef_) + fit2.intercept_)
    return stats.pearsonr(residual1, residual2)[0]


def select_variables_univariante(df, candidate_features_iv, corr_threshold=0.8, partial=True):
    """perform feature selection based on pairwise correlation coefficient
        and remove the one with smaller iv if correlation coefficient > corr_threshold
    :param df: pandas.DataFrame, with all the features transformed by woe
    :param candidate_features_iv: pd.Series with iv values for candidate features
    :param corr_threshold: threshold of correlation coefficient bigger than which
        a pair of features is considered to be correlated
    :param partial: bool, whether to use partial correlation
    :return: list of selected features
    """
    candidate_features_iv.sort_values(ascending=False, inplace=True)
    deleted_index = []
    cnt_num = len(candidate_features_iv)
    for i in range(cnt_num):
        if i in deleted_index:
            continue
        x1 = candidate_features_iv.index[i]
        for j in range(cnt_num):
            if i == j or j in deleted_index:
                continue
            y1 = candidate_features_iv.index[j]
            if partial:
                roh = calculate_partial_correlation(
                    df[candidate_features_iv.index.tolist()], x1, y1)
            else:
                roh = np.corrcoef(df[x1], df[y1])[0, 1]
            if abs(roh) > corr_threshold:
                x1_iv = candidate_features_iv[x1]
                y1_iv = candidate_features_iv[y1]
                if x1_iv > y1_iv:
                    deleted_index.append(j)
                else:
                    deleted_index.append(i)
    single_analysis_vars = [candidate_features_iv.index[i] for i
                            in range(cnt_num) if i not in deleted_index]
    return single_analysis_vars


def select_variables_multivariant_forward(x, y,
                                          candidate_features=[],
                                          initial_list=[],
                                          threshold_in=0.05,
                                          check_positive_coef=True,
                                          verbose=True):
    """ Perform a forward feature selection for logistic regression
    Arguments:
        candidate_features - list of the names the candidate features
        x - pandas.DataFrame with candidate features
        y - list-like with the target
        initial_list - list of features to start with (column names of x)
        threshold_in - include a feature if its p-value < threshold_in
        check_positive_coef - whether to make sure the coefficients are all positive
              if True, all the features should be transformed by woe
        verbose - whether to print the sequence of inclusions and exclusions
    Returns: list of selected features
    """
    if len(candidate_features) == 0:
        candidate_features = x.columns.tolist()

    included = initial_list

    # selection begins
    count = 1
    best_llr_pvalue = 10000
    while True:
        if verbose:
            print('Iteration {}'.format(count))

        count += 1

        # forward step
        excluded = list(set(candidate_features)-set(included))
        new_coef_list = []
        new_pval_list = []
        new_llr_pvalue_list = []
        new_prsquared_list = []
        new_features_list = []
        for new_column in excluded:
            model = Logit(y, add_constant(x.loc[:, included+[new_column]])).fit()
            coefficients = model.params
            pvalues = model.pvalues
            llr_pvalue = model.llr_pvalue
            prsquared = model.prsquared
            if np.isnan(pvalues.iloc[-1]):
                if verbose:
                    print('{}: bad fit'.format(new_column))
                continue
            if (check_positive_coef and (coefficients.iloc[1:] > 0).all()) \
                    or (not check_positive_coef):
                new_features_list.append(new_column)
                new_coef_list.append(coefficients)
                new_pval_list.append(pvalues)
                new_llr_pvalue_list.append(llr_pvalue)
                new_prsquared_list.append(prsquared)
        if len(new_pval_list) == 0:
            if verbose:
                print('positive coefficients can not be guaranteed')
            break
        # the same length as new_pval_list
        new_pval_array = np.array([i.iloc[-1] for i in new_pval_list])
        # so best_pval_idx is the index into new_pval_list
        best_pval_idx = new_pval_array.argmin()
        best_pval = new_pval_array[best_pval_idx]
        if best_pval < threshold_in:
            llr_pvalue = new_llr_pvalue_list[best_pval_idx]
            if (llr_pvalue < threshold_in) or (llr_pvalue <= best_llr_pvalue):
                best_llr_pvalue = min(llr_pvalue, best_llr_pvalue)
                best_feature = new_features_list[best_pval_idx]
                included.append(best_feature)
                if verbose:
                    best_prsquared = new_prsquared_list[best_pval_idx]
                    best_coefficients = new_coef_list[best_pval_idx]
                    best_pvalues = new_pval_list[best_pval_idx]
                    display_df = pd.DataFrame(columns=['coefficients', 'pvalue',
                                                       'llr_pvalue', 'pseudoR2'],
                                              index=best_coefficients.index)
                    display_df.loc[:, 'coefficients'] = best_coefficients.values
                    display_df.loc[:, 'pvalue'] = best_pvalues.values
                    display_df.iloc[0, -2] = best_llr_pvalue
                    display_df.iloc[0, -1] = best_prsquared
                    print('Add ', best_feature)
                    print(display_df)
            else:
                if verbose:
                    print('overall fitting is not significant or improving')
                break
        else:
            if verbose:
                print('coefficient of univariant is not significant')
            break

    return included


def select_variables_multivariant_backward(x, y,
                                           candidate_features=[],
                                           threshold_in=0.05,
                                           check_positive_coef=True,
                                           verbose=True):
    """ Perform a backward feature selection for logistic regression
    Arguments:
        candidate_features - list of the names the candidate features
        x - pandas.DataFrame with candidate features
        y - list-like with the target
        threshold_in - include a feature if its p-value < threshold_in
        check_positive_coef - whether to make sure the coefficients are all positive
              if True, all the features should be transformed by woe
        verbose - whether to print the sequence of inclusions and exclusions
    Returns: list of selected features
    """
    if len(candidate_features) == 0:
        candidate_features = x.columns.tolist()

    # selection begins
    count = 1
    best_llr_pvalue = 10000
    excluded = []
    included = list(set(candidate_features))
    while True:
        if verbose:
            print('Iteration {}'.format(count))

        count += 1

        # backward step
        for col in excluded:
            included.remove(col)
        model = Logit(y, add_constant(x.loc[:, included])).fit()
        coefficients = model.params
        pvalues = model.pvalues
        llr_pvalue = model.llr_pvalue
        prsquared = model.prsquared
        if verbose:
            display_df = pd.DataFrame(columns=['coefficients', 'pvalue',
                                               'llr_pvalue', 'pseudoR2'],
                                      index=coefficients.index)
            display_df.loc[:, 'coefficients'] = coefficients.values
            display_df.loc[:, 'pvalue'] = pvalues.values
            display_df.iloc[0, -2] = llr_pvalue
            display_df.iloc[0, -1] = prsquared
            print(display_df)
        if (llr_pvalue < threshold_in) or (llr_pvalue <= best_llr_pvalue):
            new_pval_array = pvalues.values[1:]
            worst_pval_idx = new_pval_array.argmax()
            worst_pval = new_pval_array[worst_pval_idx]
            worst_feature = included[worst_pval_idx]
            best_llr_pvalue = min(best_llr_pvalue, llr_pvalue)
            if worst_pval > threshold_in:
                excluded.append(worst_feature)
                if verbose:
                    print('Pop ', worst_feature)

            else:
                if verbose:
                    print('every univariant is significant')
                break
        else:
            if verbose:
                print('overall fitting significance is no longer improving')
            break

    return included


if __name__ == '__main__':
    pdf = pd.read_excel('lr_test_data.xlsx', encoding='utf-8')
    x_train = pdf.iloc[:, :-1]
    y_train = pdf.iloc[:, -1]
    # included_features = select_variables_multivariant_forward(x_train, y_train, check_positive_coef=False)
    included_features = select_variables_multivariant_backward(x_train, y_train)
    print('included_fetures', included_features)





import numpy as np
import pandas as pd
import xgboost as xgb
from xgboost.sklearn import XGBClassifier
from sklearn.model_selection import GridSearchCV, PredefinedSplit, cross_val_score
from sklearn.metrics import f1_score, accuracy_score, roc_auc_score
from bayes_opt import BayesianOptimization
import warnings
warnings.filterwarnings('ignore')


class GridSearchXGB(object):
    def __init__(self, alg):
        """
        :param alg: XGBClassifier object
          please note that the alg that is used to initialize the class instance will be changed
          after gridsearch because python uses reference of address like that in C++
        """
        self._alg = alg

    def get_params(self):
        return self._alg.get_params()

    def set_params(self, **setting_params):
        self._alg.set_params(**setting_params)

    def predict(self, x):
        """
        :param x: array-like(dim=2), Feature matrix
        """
        # np.array
        # defaults to best_ntree_limit if early stopping is available,
        # otherwise 0 (use all trees).
        prediction = self._alg.predict(x)
        return prediction

    def predict_proba(self, x):
        """
        :param x: array-like(dim=2), Feature matrix
        """
        # np.array
        # defaults to best_ntree_limit if early stopping is available,
        # otherwise 0 (use all trees).
        prediction = self._alg.predict_proba(x)
        return prediction

    def predict_proba_with_best_threshold(self, x, y, metric_list=['f1']):
        """
        tune threshold with best threshold
        :param x: array-like(dim=2), Feature matrix
        :param y: array-like(dim=1), Label
        :param metric_list: a list of metrics
        """
        max_score_list = []
        best_t_list = []
        for metric in metric_list:
            assert metric in ['f1', 'accuracy', 'roc_auc']
            max_score_list.append([metric, -1])
            best_t_list.append([metric, -1])
        y_prediction = self.predict_proba(x)[:, 1]
        for t in np.arange(0, 1, 0.05):
            y_hat = np.zeros_like(y_prediction)
            y_hat[y_prediction >= t] = 1
            for i in range(len(metric_list)):
                max_score = max_score_list[i][1]
                metric = max_score_list[i][0]
                if metric == 'f1':
                    score = f1_score(y, y_hat)
                elif metric == 'accuracy':
                    score = accuracy_score(y, y_hat)
                elif metric == 'roc_auc':
                    score = roc_auc_score(y, y_hat)
                else:
                    raise RuntimeError('not-supported metric: {}'.format(metric))
                if score > max_score:
                    max_score_list[i][1] = score
                    best_t_list[i][1] = t

        return y_prediction, max_score_list, best_t_list

    def fit(self, x_train, y_train, param_dict=None):
        if param_dict is not None:
            self._alg = XGBClassifier(**param_dict)
        self._alg.fit(x_train, y_train)

    def _custom_f1_score(self, y_pred, y_true_dmatrix):
        """
        the signature is func(y_predicted, DMatrix_y_true) where DMatrix_y_true is
        a DMatrix object such that you may need to call the get_label method.
        It must return a (str, value) pair where the str is a name for the evaluation
        and value is the value of the evaluation function.
        The callable function is always minimized.
        :param y_pred: np.array, probability score predicted by the xgbclassifier
        :param y_true_dmatrix: xgb DMatrix, true label, with positive instances as 1
        """
        y_true = y_true_dmatrix.get_label()
        y_hat = np.zeros_like(y_pred)
        y_hat[y_pred >= self._alg.get_params()['base_score']] = 1
        f1 = f1_score(y_true, y_hat)
        return 'f1_err', 1 - f1

    def _regularize_metric(self, eval_metric, early_stopping_rounds):
        """
        :param eval_metric: list of str, each str should be a built-in metric of sklearn
        :param early_stopping_rounds: int or None
        """
        feval = None
        eval_metric_list = []
        for metric in eval_metric:
            if early_stopping_rounds is not None:
                if metric == 'f1':
                    # when this happens, feval will always be used for early stopping.
                    # For custom function, the smaller the returned value the better,
                    # but for built in metrics, the cv package will automatically decide
                    # according to the metric type, e.g., the smaller the better for error
                    # but the bigger the better for auc, etc.
                    feval = self._custom_f1_score
                elif metric == 'accuracy':
                    # It is calculated as #(wrong cases)/#(all cases).
                    # The evaluation will regard the instances
                    # with prediction value larger than 0.5 as positive instances, i.e., 1 instances)
                    eval_metric_list.append('error')
                elif metric == 'roc_auc':
                    eval_metric_list.append('auc')
                else:
                    raise RuntimeError('not-supported metric: {}'.format(metric))
            else:
                if metric in ['f1', 'accuracy', 'roc_auc']:
                    eval_metric_list.append(metric)
                else:
                    raise RuntimeError('not-supported metric: {}'.format(metric))
        return feval, eval_metric_list

    def _regularize_xgb_params(self, **params_to_optimize):
        regularized_dict = {}
        for key, value in self._alg.get_params().items():
            if key in params_to_optimize.keys():
                if key in ['max_depth', 'n_estimators', 'min_child_weight',
                           'max_delta_step', 'num_parallel_tree']:
                    regularized_dict[key] = int(params_to_optimize[key])
                else:
                    regularized_dict[key] = params_to_optimize[key]
            else:
                regularized_dict[key] = value
        return regularized_dict

    def _make_train_val(self, x_train, y_train, eval_set, cv):
        """
        :param x_train: array-like(dim=2), Feature matrix
        :param y_train: array-like(dim=1), Label
        :param eval_set: a list of (X, y) tuples, only one tuple is supported as of now
        :param cv: int, number of folds for cross_validation
        """
        if eval_set is not None:
            print('using self-defined eval-set')
            assert len(eval_set) == 1
            if type(x_train) is pd.core.frame.DataFrame:
                x_train_val = pd.concat([x_train, eval_set[0][0]], axis=0)
                y_train_val = pd.concat([y_train, eval_set[0][1]], axis=0)
            else:
                x_train_val = np.concatenate((x_train, eval_set[0][0]), axis=0)
                y_train_val = np.concatenate((y_train, eval_set[0][1]), axis=0)
            # initialize all indices to 0 except the section of training
            # to -1, which means this part will not be in validation.
            # So only one fold is made
            test_fold = np.zeros(x_train_val.shape[0])
            test_fold[:x_train.shape[0]] = -1
            cv = PredefinedSplit(test_fold=test_fold)
            folds = []
            for train_indices_array, val_indices_array in cv.split():
                folds.append((train_indices_array.tolist(),
                              val_indices_array.tolist()))
        else:
            print('using cv {}'.format(cv))
            x_train_val = x_train
            y_train_val = y_train
            folds = None
        return x_train_val, y_train_val, cv, folds

    def search_with_bayesian(self, x_train, y_train, tuning_dict,
                             eval_set=None, eval_metric=['f1'], cv=3, verbose=True,
                             init_points=5, n_iter=25, acq='ucb', kappa=2.576, xi=0.0, **gp_params):
        """
        :param x_train: array-like(dim=2), Feature matrix
        :param y_train: array-like(dim=1), Label
        :param tuning_dict: dict with keys being params in XGBClassifier,
         and values being a list of lowerbound and upperbound
        :param eval_set: a list of (X, y) tuples, only one tuple is supported as of now
        :param eval_metric: a list of str
         str should be a built-in evaluation metric of sklearn, but only the last will be used
        :param cv: int, number of folds for cross_validation
        :param verbose: bool
        :param init_points: int, number of initial points for bayesian optimization
        :param n_iter: int, number of other points for bayesian optimization besides init_points
        :param acq: 'str', acquisition function used in exporation stage of bayesian,
          must be in ['ucb', 'ei', 'poi']
        :param kappa: param used in acquisition function. As of 'ucb' function, for example,
          the bigger kappas is, the more likely our bayesian is gonna search the unknown spaces
        :param xi: param used in acquisition function
        :param gp_params: param used in gaussian process
        """
        feval, eval_metric_list = self._regularize_metric(eval_metric, None)
        x_train_val, y_train_val, cv, _ = self._make_train_val(x_train, y_train, eval_set, cv)

        # use closure to pass in x_train_val and y_train_val, scoring and n_jobs, cv
        # because the params of the objective function can only be the params to optimize
        def rf_cv(**params_to_optimize):
            xgb_param = self._regularize_xgb_params(**params_to_optimize)
            # cross_val_score returns an array with each element being the val score
            # on each fold
            val = cross_val_score(XGBClassifier(**xgb_param),
                                  x_train_val, y_train_val,
                                  scoring=eval_metric_list[-1],
                                  n_jobs=self._alg.get_params()['n_jobs'],
                                  cv=cv).mean()
            return val
        rf_bo = BayesianOptimization(f=rf_cv, pbounds=tuning_dict,
                                     random_state=self._alg.get_params()['random_state'],
                                     verbose=2 if verbose else 0)
        rf_bo.maximize(init_points=init_points, n_iter=n_iter, acq=acq,
                       kappa=kappa, xi=xi, **gp_params)
        optimized_xgb_params = self._regularize_xgb_params(**rf_bo.max['params'])
        # rf_bo.maximize will not get self._alg fitted
        # please note that the alg that is used to initialize the class instance is also changed
        # because python uses reference of address like that in C++
        self._alg.set_params(**optimized_xgb_params)
        print('best tuning: {} with val score {}'.format(rf_bo.max['params'],
                                                         rf_bo.max['target']))
        print('current model: {}'.format(self._alg.get_params()))
        return self._alg.get_params()

    def search(self, x_train, y_train, tuning_dict, eval_set=None, eval_metric=['f1'], cv=3,
               early_stopping_rounds=None, verbose=True):
        """
        :param x_train: array-like(dim=2), Feature matrix
        :param y_train: array-like(dim=1), Labels
        :param tuning_dict: dict with keys being params in XGBClassifier,
         and values being a list of grid point to probe
        :param eval_set: a list of (X, y) tuples, only one tuple is supported as of now
        :param eval_metric: a list of str
         str should be a built-in evaluation metric of sklearn.
        :param cv: int, number of folds for cross_validation
        :param early_stopping_rounds: int, Validation metric needs to improve at least once
         in every 'early_stopping_rounds' round(s) to continue training.
         Note that we use this parameter for the search of best number of estimators so all the other
         parameters will be kept unchanged
         If there’s more than one metric in eval_metric, the last metric will be used for early stopping
        :param verbose: bool, If verbose, prints the evaluation metric measured on the validation set.
        """
        feval, eval_metric_list = self._regularize_metric(eval_metric, early_stopping_rounds)
        if early_stopping_rounds is not None:
            print('warning: only the number of estimators will be increased '
                  'until early stopping criteria'
                  'is met and the tuning grid will not be used')
            x_train_val, y_train_val, cv, folds = self._make_train_val(x_train, y_train, eval_set, cv)
            xgb_dmatrix = xgb.DMatrix(x_train_val, label=y_train_val)
            xgb_param = self._alg.get_xgb_params()
            # when folds is defined in xgb.cv, nfold will not be used
            cvresult = xgb.cv(xgb_param, xgb_dmatrix,
                              num_boost_round=self._alg.get_params()['n_estimators'],
                              nfold=cv,
                              metrics=eval_metric_list,
                              feval=feval,
                              folds=folds,
                              early_stopping_rounds=early_stopping_rounds,
                              verbose_eval=True if verbose else None,
                              seed=self._alg.get_params()['random_state'])
            # returns the best n_estimators
            # please note that the alg that is used to initialize the class instance is also changed
            # because python uses reference of address like that in C++
            self._alg.set_params(n_estimators=cvresult.shape[0])
            print('best n_estimators: {} with val score {}'.format(self._alg.get_params()['n_estimators'],
                                                                   cvresult.iloc[-1, :][-2]))
            print('current model: {}'.format(self._alg.get_params()))
            return self._alg.get_params()
        else:
            x_train_val, y_train_val, cv, _ = self._make_train_val(x_train, y_train, eval_set, cv)
            grid_search_params = {'estimator': self._alg,
                                  'param_grid': tuning_dict,
                                  'scoring': eval_metric_list,
                                  'cv': cv,
                                  'verbose': 2 if verbose else 0,
                                  'refit': eval_metric_list[-1],
                                  'n_jobs': self._alg.get_params()['n_jobs']}
            grsearch = GridSearchCV(**grid_search_params)
            grsearch.fit(x_train_val, y_train_val)
            # after grsearch-fitting, the estimator in GridSearchCV will not get fitted
            # please note that the alg that is used to initialize the class instance is also changed
            # because python uses reference of address like that in C++
            self._alg.set_params(**grsearch.best_params_)
            print('best tuning: {} with val score {}'.format(grsearch.best_params_,
                                                             grsearch.best_score_))
            print('current model: {}'.format(self._alg.get_params()))
            return self._alg.get_params()


def test_tunning_grid():
    from sklearn.datasets import make_classification
    from sklearn.model_selection import train_test_split
    print('preparing data...')
    # 产生随机分类数据集，10个特征， 2个类别
    x, y = make_classification(n_samples=1000, n_features=10, n_classes=2, random_state=2)
    train_cols_list = ['F1', 'F2', 'F3', 'F4', 'F5', 'F6', 'F7', 'F8', 'F9', 'F10']
    pdf = pd.DataFrame(x, columns=train_cols_list)
    pdf['label'] = y
    x_train, x_test, y_train, y_test = train_test_split(pdf.loc[:, train_cols_list],
                                                        pdf.loc[:, 'label'],
                                                        random_state=1)
    alg = XGBClassifier(learning_rate=0.1,
                        n_estimators=1000,
                        max_depth=5,
                        min_child_weight=1,
                        gamma=0,
                        subsample=0.8,
                        colsample_bytree=0.8,
                        objective='binary:logistic',
                        reg_lambda=1,
                        reg_alpha=0,
                        n_jobs=4,
                        scale_pos_weight=1,
                        random_state=27)
    tuning_grid = {'gamma': [0.1, 0.2]}
    gs = GridSearchXGB(alg)
    best_param = gs.search(x_train, y_train, tuning_grid)
    print('search result')
    print(best_param)
    print('current result')
    print(gs.get_params())


def test_early_stopping():
    from sklearn.datasets import make_classification
    from sklearn.model_selection import train_test_split
    print('preparing data...')
    # 产生随机分类数据集，10个特征， 2个类别
    x, y = make_classification(n_samples=1000, n_features=10, n_classes=2, random_state=2)
    train_cols_list = ['F1', 'F2', 'F3', 'F4', 'F5', 'F6', 'F7', 'F8', 'F9', 'F10']
    pdf = pd.DataFrame(x, columns=train_cols_list)
    pdf['label'] = y
    x_train, x_test, y_train, y_test = train_test_split(pdf.loc[:, train_cols_list],
                                                        pdf.loc[:, 'label'],
                                                        random_state=1)
    alg = XGBClassifier(learning_rate=0.1,
                        n_estimators=1000,
                        max_depth=5,
                        min_child_weight=1,
                        gamma=0,
                        subsample=0.8,
                        colsample_bytree=0.8,
                        objective='binary:logistic',
                        reg_lambda=1,
                        reg_alpha=0,
                        n_jobs=4,
                        scale_pos_weight=1,
                        random_state=27)
    tuning_grid = {'gamma': [0.1, 0.2]}
    gs = GridSearchXGB(alg)
    best_param = gs.search(x_train, y_train, tuning_grid,
                           [(x_test, y_test)], ['f1', 'roc_auc'],
                           early_stopping_rounds=3)
    print('search result')
    print(best_param)
    print('recheck current result')
    print(gs.get_params())


def test_consecutive_tuning_and_predict():
    from sklearn.datasets import make_classification
    from sklearn.model_selection import train_test_split
    print('preparing data...')
    # 产生随机分类数据集，10个特征， 2个类别
    x, y = make_classification(n_samples=1000, n_features=10, n_classes=2, random_state=2)
    train_cols_list = ['F1', 'F2', 'F3', 'F4', 'F5', 'F6', 'F7', 'F8', 'F9', 'F10']
    pdf = pd.DataFrame(x, columns=train_cols_list)
    pdf['label'] = y
    x_train, x_test, y_train, y_test = train_test_split(pdf.loc[:, train_cols_list],
                                                        pdf.loc[:, 'label'],
                                                        random_state=1)
    alg = XGBClassifier(learning_rate=0.1,
                        n_estimators=1000,
                        max_depth=5,
                        min_child_weight=1,
                        gamma=0,
                        subsample=0.8,
                        colsample_bytree=0.8,
                        objective='binary:logistic',
                        reg_lambda=1,
                        reg_alpha=0,
                        n_jobs=4,
                        scale_pos_weight=1,
                        random_state=27)
    tuning_grid = {'gamma': [0.1, 0.2]}
    gs = GridSearchXGB(alg)
    best_param = gs.search(x_train, y_train, tuning_grid,
                           eval_set=[(x_test, y_test)], eval_metric=['accuracy', 'roc_auc'],
                           early_stopping_rounds=3)
    print('after n_estimators search is completed...')
    print(gs.get_params())
    best_param = gs.search(x_train, y_train, tuning_grid,
                           eval_set=[(x_test, y_test)], eval_metric=['accuracy', 'roc_auc'])
    print('after gamma_search is completed...')
    print(gs.get_params())
    gs.fit(x_train, y_train)
    print('model after fit...')
    print(gs.get_params())
    prediction, best_score_list, best_threshold_list = \
        gs.predict_proba_with_best_threshold(x_test, y_test, ['f1', 'roc_auc'])
    print(best_score_list)
    print(best_threshold_list)


def test_bayesian():
    from sklearn.datasets import make_classification
    from sklearn.model_selection import train_test_split
    print('preparing data...')
    # 产生随机分类数据集，10个特征， 2个类别
    x, y = make_classification(n_samples=1000, n_features=10, n_classes=2, random_state=2)
    train_cols_list = ['F1', 'F2', 'F3', 'F4', 'F5', 'F6', 'F7', 'F8', 'F9', 'F10']
    pdf = pd.DataFrame(x, columns=train_cols_list)
    pdf['label'] = y
    x_train, x_test, y_train, y_test = train_test_split(pdf.loc[:, train_cols_list],
                                                        pdf.loc[:, 'label'],
                                                        random_state=1)
    alg = XGBClassifier(learning_rate=0.1,
                        n_estimators=1000,
                        max_depth=5,
                        min_child_weight=1,
                        gamma=0,
                        subsample=0.8,
                        colsample_bytree=0.8,
                        objective='binary:logistic',
                        reg_lambda=1,
                        reg_alpha=0,
                        n_jobs=4,
                        scale_pos_weight=1,
                        random_state=27)
    tuning_grid = {'learning_rate': [0.11, 0.3], 'n_estimators': [10, 200]}
    gs = GridSearchXGB(alg)
    best_param = gs.search_with_bayesian(x_train, y_train, tuning_dict=tuning_grid,
                                         eval_set=[(x_test, y_test)], eval_metric=['f1', 'roc_auc'],
                                         init_points=1, n_iter=3)
    print('after bayesian search is completed...')
    print(gs.get_params())


if __name__ == '__main__':
    test_consecutive_tuning_and_predict()


import numpy as np
from sklearn.model_selection import StratifiedKFold
from itertools import product


class MeanEncoder():
    def __init__(self, categorical_features, n_splits=5,
                 target_type='classification', prior_weight_func=None):
        """
        :param categorical_features: list of str,
        the name of the categorical columns to encode

        :param n_splits: the number of splits used in mean encoding

        :param target_type: str, 'regression' or 'classification'

        :param prior_weight_func:
               a function that takes in the number of observations,
               and outputs prior weight
               when a dict is passed,
               the default exponential decay function will be used:
        k: the number of observations
           needed for the posterior to be weighted
           equally as the prior
        f: larger f --> smaller slope
        """
        self.categorical_features = categorical_features
        self.n_splits = n_splits
        self.learned_stats = {}

        if target_type == 'classification':
            self.target_type = target_type
            self.target_values = []
        else:
            self.target_type = 'regression'
            self.target_values = None

        if isinstance(prior_weight_func, dict):
            self.prior_weight_func = \
                eval('lambda x: 1 / (1 + np.exp((x - k) / f))',
                     dict(prior_weight_func, np=np))
        elif callable(prior_weight_func):
            self.prior_weight_func = prior_weight_func
        else:
            self.prior_weight_func = \
                lambda x: 1 / (1 + np.exp((x - 2) / 1))

    @staticmethod
    def mean_encode_subroutine(x_train, y_train, x_test,
                               variable, target, prior_weight_func):
        x_train = x_train[[variable]].copy()
        x_test = x_test[[variable]].copy()

        if target is not None:
            nf_name = '{}_pred_{}'.format(variable, target)
            # classification
            x_train['pred_temp'] = int(y_train == target)
        else:
            nf_name = '{}_pred'.format(variable)
            # regression
            x_train['pred_temp'] = y_train
        prior = x_train['pred_temp'].mean()

        col_avg_y = x_train.groupby(by=variable, axis=0)['pred_temp'].agg(
            {'mean': 'mean', 'beta': 'size'})
        col_avg_y['beta'] = prior_weight_func(col_avg_y['beta'])
        col_avg_y[nf_name] = col_avg_y['beta'] * prior + \
                             (1 - col_avg_y['beta']) * col_avg_y['mean']
        col_avg_y.drop(['beta', 'mean'], axis=1, inplace=True)

        nf_train = x_train.join(col_avg_y, on=variable)[nf_name].values
        nf_test = x_test.join(col_avg_y, on=variable).\
            fillna(prior, inplace=False)[nf_name].values

        return nf_train, nf_test, prior, col_avg_y

    def fit_transform(self, x, y):
        """
        :param x: pandas DataFrame, n_samples * n_features
        :param y: pandas Series or numpy array, n_samples
        :return x_new: the transformed pandas DataFrame
                       containing mean-encoded categorical features
        """
        x_new = x.copy()
        skf = StratifiedKFold(self.n_splits)

        if self.target_type == 'classification':
            self.target_values = sorted(set(y))
            self.learned_stats = {
                '{}_pred_{}'.format(variable, target): []
                for variable, target in
                product(self.categorical_features, self.target_values)}
            for variable, target in product(self.categorical_features,
                                            self.target_values):
                nf_name = '{}_pred_{}'.format(variable, target)
                x_new.loc[:, nf_name] = np.nan
                for large_ind, small_ind in skf.split(y, y):
                    nf_large, nf_small, prior, col_avg_y = \
                        MeanEncoder.mean_encode_subroutine(
                            x_new.iloc[large_ind],
                            y.iloc[large_ind],
                            x_new.iloc[small_ind],
                            variable, target,
                            self.prior_weight_func)
                    x_new.iloc[small_ind, -1] = nf_small
                    self.learned_stats[nf_name].append((prior, col_avg_y))
        else:
            self.learned_stats = {
                '{}_pred'.format(variable): [] for variable in
                self.categorical_features}
            for variable in self.categorical_features:
                nf_name = '{}_pred'.format(variable)
                x_new.loc[:, nf_name] = np.nan
                for large_ind, small_ind in skf.split(y, y):
                    nf_large, nf_small, prior, col_avg_y = \
                        MeanEncoder.mean_encode_subroutine(
                            x_new.iloc[large_ind],
                            y.iloc[large_ind],
                            x_new.iloc[small_ind],
                            variable, None, self.prior_weight_func)
                    x_new.iloc[small_ind, -1] = nf_small
                    self.learned_stats[nf_name].append((prior, col_avg_y))
        return x_new

    def transform(self, x):
        """
        :param x: pandas DataFrame, n_samples * n_features
        :return x_new: the transformed pandas DataFrame
                       containing mean-encoded categorical features
        """
        x_new = x.copy()

        if self.target_type == 'classification':
            for variable, target in product(self.categorical_features,
                                            self.target_values):
                nf_name = '{}_pred_{}'.format(variable, target)
                x_new[nf_name] = 0
                for prior, col_avg_y in self.learned_stats[nf_name]:
                    x_new[nf_name] += \
                        x_new[[variable]].join(
                            col_avg_y,
                            on=variable).fillna(prior, inplace=False)[nf_name]
                x_new[nf_name] /= self.n_splits
        else:
            for variable in self.categorical_features:
                nf_name = '{}_pred'.format(variable)
                x_new[nf_name] = 0
                for prior, col_avg_y in self.learned_stats[nf_name]:
                    x_new[nf_name] += \
                        x_new[[variable]].join(
                            col_avg_y,
                            on=variable).fillna(prior, inplace=False)[nf_name]
                x_new[nf_name] /= self.n_splits

        return x_new



import numpy as np
import pandas as pd
from sklearn.metrics import roc_curve, auc
import matplotlib as mpl
import matplotlib.pyplot as plt
import itertools
import datetime
import os


def cutoffpoints_to_interval(cut_off_points):
    """
    一定是连续值才有所谓分割区间
    :param cut_off_points: list, 不含起止点
    :return: list, 以cut_off_points为分割点，下限为-inf,上限为inf的区间列表
    """
    cut_off_points = list(set(cut_off_points))
    cut_off_points.sort()
    group_interval = [pd.Interval(left=float('-inf'),
                                  right=float(cut_off_points[0]))]
    for i in range(len(cut_off_points)-1):
        group_interval.append(pd.Interval(left=float(cut_off_points[i]),
                                          right=float(cut_off_points[i+1])))
    group_interval.append(pd.Interval(left=float(cut_off_points[-1]),
                                      right=float('inf')))
    return group_interval


def cutoffpoints_to_edge(x, cut_off_points):
    """
    :param x: 要映射的点
    :param cut_off_points: list, 不含起止点
    :return: 映射到的点
    """
    num_of_points = len(cut_off_points)
    first_point = min(cut_off_points)
    last_point = max(cut_off_points)
    if x <= first_point:
        return first_point
    elif x > last_point:
        return 10e10
    else:
        for i in range(num_of_points - 1):
            if cut_off_points[i] < x <= cut_off_points[i+1]:
                return cut_off_points[i+1]


def map_to_inf(regroup, col):
    """
    用于将连续型分箱表的col字段取值范围的最小值映射为-inf
    将最大值映射为inf
    :param regroup:分箱表dataframe，index必须是从0开始逐渐编号
    :param col:要调整上下限的字段名string
    :return: dataframe
    """
    regroup.sort_values(col, ascending=True, inplace=True)
    regroup = regroup.astype({col: 'object'})
    # 修改上下限为-inf  inf
    regroup.loc[0, col] = \
        pd.Interval(left=float('-inf'),
                    right=regroup.loc[0, col].right)
    last_index = regroup.shape[0] - 1
    regroup.loc[last_index, col] = pd.Interval(
        left=regroup.loc[last_index, col].left,
        right=float('inf'))
    return regroup


def map_to_bin(df, from_col, bin_df, to_col, test_indicator=False,
               target=None, fill_na=None, failure_value=0):
    """
    将原数据框df的某个列from_col映射为分箱表bin_df的to_col
    :param df: pd.Dataframe, 需要映射的表
    :param from_col: str, 需要映射的df的一个列
    :param bin_df: 用于将from_col的各个取值按bin_range映射为to_col取值的dataframe
                   bin_range列为col连续区间interval或取值list
    :param to_col: str, bin_df的一个列名
    :param test_indicator: bool, if True, a new bin_df will be returned
    :param target: str, 需要映射的df的目标列
    :param fill_na: float, 缺失值填充
    :param failure_value: float, 如果最后无法映射，就按这个值映射
    :return: list, 值为映射到的值
    """
    num_values = df.shape[0]
    num_of_bins = bin_df.shape[0]
    input_series = df[from_col]
    if fill_na is not None:
        input_series = input_series.fillna(fill_na)
    # pandas里面，如果某一列dtype为数值型，则none，np.nan都作为np.nan
    # 但是如果某一列为object类型，则none和np.nan是不一样的
    # 如果map传入的是一个字典，则字典无法涉及的部分会始终映射为np.nan
    mapped_value_list = [failure_value] * num_values
    if test_indicator:
        bin_total_list = [0] * num_of_bins
        bin_bad_list = [0] * num_of_bins
    for i in range(num_values):
        value = input_series.iloc[i]
        for j in range(num_of_bins):
            if value in bin_df['bin_range'].iloc[j]:
                mapped_value_list[i] = bin_df[to_col].iloc[j]
                if test_indicator:
                    bin_total_list[j] += 1
                    if target is not None:
                        bin_bad_list[j] += df[target].iloc[i]
                break
    if test_indicator:
        if target is not None:
            bin_df2 = pd.DataFrame({'bin_range': bin_df['bin_range'],
                                    'bin_total': bin_total_list,
                                    'bin_bad': bin_bad_list},
                                   index=np.arange(num_of_bins))
            bin_df2['bin_good'] = bin_df2['bin_total'] - bin_df2['bin_bad']
            bin_df2['bin_bad_rate'] = bin_df2['bin_bad'] / bin_df2['bin_total']
            bin_df2 = compute_woe_iv(bin_df2, 'bin_good', 'bin_bad', 'bin_total')
        else:
            bin_df2 = pd.DataFrame({'bin_range': bin_df['bin_range'],
                                    'bin_total': bin_total_list},
                                   index=np.arange(num_of_bins))
        # 在最后两列生成feature_name和bin_no
        bin_df2['feature_name'] = from_col
        bin_df2['bin_no'] = bin_df['bin_no']
        # 将最后两列调整到第前面
        bin_df2 = pd.concat(
            [bin_df2[['feature_name', 'bin_no']], bin_df2.iloc[:, :-2]], axis=1)
        return mapped_value_list, bin_df2
    else:
        return mapped_value_list


def unsupervised_split(df, var, num_of_split=5, quantile=True):
    """
    :param df: 数据集
    :param var: 需要分箱的变量, 仅限数值型
    :param num_of_split: 需要分箱个数，默认是5
    :param quantile: bool 默认是等频，否则是等距
    :return: list of split points, 不含首尾点
    """
    if quantile:
        interval_index = int(df.shape[0] / num_of_split)
        split_point_indices = [i * interval_index for i in range(1, num_of_split)]
        raw_values = df[var].sort_values(na_position='first')
        split_points = [raw_values.iloc[i] for i in split_point_indices]
        split_points = sorted(list(set(split_points)))
    else:
        var_max, var_min = max(df[var]), min(df[var])
        interval_value = (var_max - var_min) * 1.0 / num_of_split
        split_points = [var_min + i * interval_value for i in range(1, num_of_split)]
    return split_points


def concat_of_spec_bins(spec_bin_df, bin_df, range_col, continuous_flag):
    """
    将特殊取值对应的分箱（一定是离散型的）和已有分箱表连接起来
    :param spec_bin_df: 特殊取值对应的离散型分箱表dataframe
    :param bin_df: 已有分箱表dataframe
    :param range_col: 要连接并调整的分箱区间的字段名string
    :param continuous_flag: 已有分箱表为连续分箱表还是离散分箱表
                            如果为连续的，则拼接后的表仍然是连续的,
                                range_col字段存储的是col_values连续区间
                                取值为interval类型
                            如果是离散的，则拼接后的表仍然是离散的，
                                range_col字段存储的是col_values的取值list
                                取值为list类型
    :return: 分箱表dataframe
    """
    regroup = bin_df.copy()
    regroup_single = spec_bin_df.copy()
    if not continuous_flag:
        regroup_single[range_col] = regroup_single[range_col].map(lambda x: [x])
    # 对于连续型变量，如果有缺失值需要单独设置，一定要人为设置这个缺失值比所有其他数值都小
    regroup = pd.concat([regroup_single, regroup], sort=False)
    if continuous_flag:
        group_interval = regroup[range_col].tolist()
        # 取每个区间的最大值, cut_off_points不含起止点
        cut_off_points = []
        for i in group_interval[:-1]:
            if isinstance(i, pd.Interval):
                cut_off_points.append(i.right)
            else:
                cut_off_points.append(i)
        # 第一个区间就是(-inf---特殊值], 第二个区间为（特殊值----某个值].....
        regroup[range_col] = cutoffpoints_to_interval(cut_off_points)
    # 重新把index编号调整为从0开始逐渐增大
    regroup.reset_index(drop=True, inplace=True)
    return regroup


# def categorical_values_to_bins(categorical_values_df, col_key, col_values):
#     """
#     :param categorical_values_df: 数据框dataframe，必须含col_key, col_values字段
#     :param col_key: string, 字段名，该字段存储了区间划分的依据
#     :param col_values: string, 字段名，该字段存储了非重复的类别取值
#     :return: bin_range_df, 只有一个col: bin_range, 取值为分箱对应的col_values取值list
#     """
#     col_values_indices_dict = categorical_values_df.groupby(col_key).indices
#     bin_range_df = pd.DataFrame(columns=['bin_range'], dtype='object')
#     for key_range, indices in col_values_indices_dict.items():
#         values = categorical_values_df.loc[indices, col_values].tolist()
#         values = values if isinstance(values, list) else [values]
#         bin_df = pd.DataFrame(columns=['bin_range'], dtype='object')
#         bin_df.at[0, 'bin_range'] = values
#         bin_range_df = pd.concat([bin_range_df, bin_df])
#     bin_range_df.reset_index(drop=True, inplace=True)
#     return bin_range_df


# def categorical_bins_to_values(categorical_bins_df, col_no, col_range):
#     """
#     :param categorical_bins_df: 数据框dataframe，必须含col_no, col_bins字段
#     :param col_no: string, 分箱号的字段名，取值是非重复的
#     :param col_range: string, 字段名，存储了每个分箱对应的col_values取值列表
#     :return: binno_to_colvalues_dict, 键值为bin_no，取值为分箱对应的col_values取值list
#              binno_to_colvalues_df, 主要是为了便于展示分箱结果
#     """
#     binno_to_colvalues_df = pd.DataFrame(columns=['bin_range', 'bin_no'])
#     binno_to_colvalues_dict = {}
#     binrange_list = categorical_bins_df(col_range).tolist()
#     binno_list = categorical_bins_df(col_no).tolist()
#     for bin_no, col_values in zip(binno_list, binrange_list):
#         col_values = col_values if isinstance(col_values, list) else [col_values]
#         binno_to_colvalues_dict[bin_no] = col_values
#         bin_df = pd.DataFrame(columns=['bin_range', 'bin_no'])
#         bin_df['bin_range'] = col_values
#         bin_df['bin_no'] = bin_no
#         binno_to_colvalues_df = pd.concat([binno_to_colvalues_df, bin_df])
#     binno_to_colvalues_df.reset_index(drop=False, inplace=True)
#     binno_to_colvalues_df = \
#         binno_to_colvalues_df.astype({'bin_range': 'float', 'bin_no': 'object'})
#     return binno_to_colvalues_dict, binno_to_colvalues_df


def cal_chi2(df, bad_col, good_col, total_col):
    """
    :param df: 只有两个分箱的样本数据.
    :param bad_col: 列明string, 统计某个取值的坏样本数
    :param good_col: 列明string, 统计某个取值的好样本数
    :param total_col: 列明string, 统计某个取值的全部样本数
    :return: chi2 二联卡卡方值
    """
    all_bad_rate = df[bad_col].sum() * 1.0 / df[total_col].sum()
    all_good_rate = df[good_col].sum() * 1.0 / df[total_col].sum()
    # 当全部样本只有好或者坏样本时，卡方值为0
    if all_bad_rate in [0, 1]:
        return 0.0
    df2 = df.copy()
    # 计算每组的坏用户期望数量
    df2['bad_expected'] = df2[total_col] * all_bad_rate
    df2['good_expected'] = df2[total_col] * all_good_rate
    # 遍历每组的坏用户期望数量和实际数量
    bad_combined = zip(df2['bad_expected'], df2[bad_col])
    good_combined = zip(df2['good_expected'], df2[good_col])
    # 计算每组的卡方值
    bad_chi = [(i[0] - i[1]) ** 2 / (i[0] + 1e-6) for i in bad_combined]
    good_chi = [(i[0] - i[1]) ** 2 / (i[0] + 1e-6) for i in good_combined]
    # 计算总的卡方值
    chi2 = sum(bad_chi) + sum(good_chi)
    return chi2


def combine_bin_df(regroup, col, best_combined_index, continuous_flag=False):
    """
    :param regroup: dataframe [feature_name(optional), col, bin_total] or
      [feature_name(optional), col, bin_total, bin_bad, bin_good, bin_bad_rate]
      each row of col is a list if continuous_flag is False, else a pd.Interval.
      the indices must be increasing from 0 with increment of 1
    :param col: string, name of the feature range
    :param best_combined_index: int, the rows of this index
      and the next index of regroup will be combined and then filled
      into the row of this index, and then the row of the next index will be removed
    :param continuous_flag: bool
    """
    regroup_df = regroup.copy()
    # do this because pd.categorical dtype is not allowed to edit
    if continuous_flag:
        regroup_df[[col]] = regroup_df[[col]].astype('object')
    combine_df = regroup_df.loc[best_combined_index:best_combined_index+1, :]
    if continuous_flag:
        # pd.dataframe.at is like loc,
        # must use the index and column name, rather than positions
        regroup_df.at[best_combined_index, col] = pd.Interval(
            left=combine_df[col][best_combined_index].left,
            right=combine_df[col][best_combined_index+1].right)
    else:
        regroup_df.at[best_combined_index, col] = \
            combine_df[col][best_combined_index] + combine_df[col][best_combined_index+1]
    regroup_df.at[best_combined_index, 'bin_total'] = combine_df['bin_total'].sum()
    if 'bin_bad_rate' in regroup.columns.tolist():
        regroup_df.at[best_combined_index, 'bin_bad'] = combine_df['bin_bad'].sum()
        regroup_df.at[best_combined_index, 'bin_good'] = combine_df['bin_good'].sum()
        regroup_df.at[best_combined_index, 'bin_bad_rate'] = \
            regroup_df['bin_bad'][best_combined_index] / regroup_df['bin_total'][best_combined_index]
    # 删除合并前的右区间, regroup_df的index始终保持从0开始从小到大排序
    regroup_df = regroup_df.loc[regroup_df.index != (best_combined_index+1), :]
    # make sure the indices are increasing from 0 with increment 1 when bin_df's returned
    regroup_df.reset_index(drop=True, inplace=True)
    return regroup_df


def bin_bad_rate(df, col, target):
    """
    :param df: dataframe
    :param col: string, 需要计算好坏率的变量名
    :param target:string, 目标变量的字段名
    :return:dataframe[col, bin_total, bin_bad, bin_good, bin_bad_rate]
            按照col取值去重后从小到大排列， index一定是从0开始逐渐编号
    """
    # 按col的值去重后从小到大排列，并作为index
    total = df.groupby([col])[target].count()
    bad = df.groupby([col])[target].sum()
    regroup = pd.merge(pd.DataFrame({'bin_total': total}),
                       pd.DataFrame({'bin_bad': bad}),
                       left_index=True, right_index=True, how='left')
    # 默认drop=false，原来的index，即col的值去重后从小到大排列的值，另外生成一列
    regroup = regroup.reset_index()
    # 计算根据col分组后每组的违约率, 对于无顺序型数据，需要计算好坏比来代替原来离散的数值
    regroup['bin_good'] = regroup['bin_total'] - regroup['bin_bad']
    regroup['bin_bad_rate'] = regroup['bin_bad'] / regroup['bin_total']
    return regroup


def compute_woe_iv(bin_df, col_good, col_bad, col_total):
    """
    param bin_df:DataFrame|分箱表, 按照bin_range,
        或col_values取值去重后从小到大排列
    param col_good:str|feature名称，统计每个分箱的好样本个数
    param col_bad:str|feature名称，统计每个分箱的坏样本个数
    param col_total:str|feature名称，统计每个分箱一共有多少样本
    return: 原dataframe增添多列
    """
    d2 = bin_df.copy()
    total = d2[col_total].sum()
    bad = d2[col_bad].sum()
    good = total - bad

    d2['bin_good_rate'] = d2[col_good] / d2[col_total]
    d2['badattr'] = d2[col_bad] / bad
    d2['goodattr'] = d2[col_good] / good
    sum_bad = 0
    acc_bad_rate_list = []
    for i in d2[col_bad]:
        sum_bad += i
        acc_bad_rate_list.append(sum_bad / bad)
    d2['acc_bad_rate'] = acc_bad_rate_list
    sum_good = 0
    acc_good_rate_list = []
    for i in d2[col_good]:
        sum_good += i
        acc_good_rate_list.append(sum_good / good)
    d2['acc_good_rate'] = acc_good_rate_list
    d2['bin_total_rate'] = d2[col_total] / total
    d2['total_bad_rate'] = bad / total

    def compute_woe(badattr, goodattr):
        return np.log(badattr / (goodattr + 1e-6) + 1e-6)
    # 如果有全bad的分箱，则该分箱woe=inf，表示该分箱的坏账率远大于整体坏账率
    # 如果有全good的分箱，则woe=-inf，表示该分箱坏账率远小于整体坏账率
    # 如果分箱的坏账率和整体一致，则woe=0
    # 因此woe取值范围是(-inf---inf)
    d2['bin_woe'] = d2.apply(
        lambda x: compute_woe(x['badattr'], x['goodattr']),
        axis=1)
    # 如果有全bad或全good的分箱，则iv=inf，表示该分箱的坏账率水平和整体相比差异大
    #   该分箱预测能力极高
    # 如果分箱坏账率和整体一致，则iv=0，表示该分箱没有预测能力
    # 因此iv取值范围是[0----inf)
    d2['bin_iv'] = d2.apply(
        lambda x: (x['badattr'] - x['goodattr']) * x['bin_woe'],
        axis=1)
    # 变量的总iv
    d2['iv'] = d2['bin_iv'].sum()

    d2.drop(['badattr', 'goodattr'], axis=1, inplace=True)

    return d2


def cal_ks(predict, target, plot=False):
    """
    ks经济学意义:
      将预测为坏账的概率从大到小排序，然后按从大到小依次选取一个概率值作为阈值，
      大于阈值的部分为预测为坏账的部分--记录其中真实为坏账的个数， 真实为好账的个数，
      上述记录值每次累加且除以总的坏账个数即累计坏账率，除以总好账个数为累计好账率, 累加结果存入列表
    sklearn.metrics.roc_curve（二分类标签，预测为正例的概率或得分）:
      将预测为正例（默认为1）的概率（0-1间）或得分（不限大小）从大到小排序, 然后按从大到小依次选取一个值作为阈值
      大于阈值的部分为预测为正例的部分--其中真实为正例的个数即TP, 真实为负例的个数即为FP
      上述值每次累加且除以总的正例个数为TPR, 除以总的负例个数为FPR，累加结果存入列表
    ks = max(累计坏账率list - 累计好账率list) = max(TPR_list - FPR_list)
    :param predict: list like, 可以为某个数值型特征字段，也可以是预测为坏账的概率的字段
    :param target: list like, 好坏账标签字段，字段中1为坏账
    :param plot: bool, 是否画图
    :return: ks, ks_thresh
    """
    # fpr即FPR_list, tpr即TPR_list, thresholds为上述所谓依次选取的阈值
    # thresholds一定是递减的，第一个值为max(预测为正例的概率或得分)+1
    fpr, tpr, thresholds = roc_curve(target, predict)
    ks = (tpr-fpr).max()
    ks_index = np.argmax(tpr-fpr)
    ks_thresh = thresholds[ks_index]
    if plot:
        # 绘制曲线
        plt.plot(tpr, label='bad_cum', linewidth=2)
        plt.plot(fpr, label='good_cum', linewidth=2)
        plt.plot(tpr-fpr, label='ks_curve', linewidth=2)
        # 标记ks点
        x_point = (ks_index, ks_index)
        y_point = (fpr[ks_index], tpr[ks_index])
        plt.plot(x_point, y_point, label='ks {:.2f}@{:.2f}'.format(ks, ks_thresh),
                 color='r', marker='o', markerfacecolor='r',
                 markersize=5)
        plt.scatter(x_point, y_point, color='r')
        # 绘制x轴（阈值）, thresholds第一个值为max(预测为正例的概率或得分)+1, 因此不画出来
        effective_indices_num = thresholds[1:].shape[0]
        if effective_indices_num > 5:
            # 向下取整
            increment = int(effective_indices_num / 5)
        else:
            increment = 1
        indices = range(1, thresholds.shape[0], increment)
        plt.xticks(indices, [round(i, 2) for i in thresholds[indices]])
        plt.xlabel('thresholds')
        plt.legend()
        plt.show()
    return ks, ks_thresh


def plot_roc(label, prediction):
    fpr, tpr, thresholds = roc_curve(label, prediction, pos_label=1)
    metric = auc(fpr, tpr)
    plt.plot(fpr, tpr)
    plt.xlim([0.0, 1.0])
    plt.ylim([0.0, 1.0])
    plt.title('ROC curve, auc:%.4f' % metric)
    plt.xlabel('False Positive Rate (1 - Specificity)')
    plt.ylabel('True Positive Rate (Sensitivity)')
    plt.grid(True)
    plt.show()


def calc_lift(df, pred, target, groupnum=None, plot=True):
    if groupnum is None:
        groupnum = len(df.index)

    def n0(x): return sum(x == 0)
    def n1(x): return sum(x == 1)
    dfkslift = df.sort_values(pred, ascending=False).reset_index(drop=True)\
        .assign(group=lambda x: np.ceil((x.index+1)/(len(x.index)/groupnum)))\
        .groupby('group')[target].agg([n0, n1])\
        .reset_index().rename(columns={'n0': 'good', 'n1': 'bad'})\
        .assign(
        group=lambda x: (x.index+1)/len(x.index),
        good_distri=lambda x: x.good/sum(x.good),
        bad_distri=lambda x: x.bad/sum(x.bad),
        badrate=lambda x: x.bad/(x.good+x.bad),
        cumbadrate=lambda x: np.cumsum(x.bad)/np.cumsum(x.good+x.bad),
        lift=lambda x: (np.cumsum(x.bad)/np.cumsum(x.good+x.bad))/(sum(x.bad)/sum(x.good+x.bad)),
        cumgood=lambda x: np.cumsum(x.good)/sum(x.good),
        cumbad=lambda x: np.cumsum(x.bad)/sum(x.bad))\
        .assign(ks=lambda x: abs(x.cumbad-x.cumgood))
    dfkslift = pd.concat([pd.DataFrame({'group': 0, 'good': 0, 'bad': 0,
                                        'good_distri': 0, 'bad_distri': 0,
                                        'badrate': 0, 'cumbadrate': np.nan,
                                        'cumgood': 0, 'cumbad': 0, 'ks': 0, 'lift': np.nan}, index=np.arange(1)),
                          dfkslift], ignore_index=True)

    if plot:
        dfks = dfkslift.loc[lambda x: x.ks == max(x.ks)].sort_values('group').iloc[0]
        fig, ax = plt.subplots()
        ax.plot(dfkslift.group, dfkslift.ks, 'b-',
                 dfkslift.group, dfkslift.cumgood, 'k-',
                 dfkslift.group, dfkslift.cumbad, 'k-')
        ax.plot([dfks['group'], dfks['group']], [0, dfks['ks']], 'r--')
        fig.gca().set(title='K-S', xlabel='worst % of population', ylabel='total good/bad rate',
                      xlim=[0, 1], ylim=[0, 1], aspect='equal')
        ax.text(0.2, 0.8, 'Bad', horizontalalignment='center')
        ax.text(0.8, 0.55, 'Good', horizontalalignment='center')
        ax.text(dfks['group'], dfks['ks'], 'KS:' + str(round(dfks['ks'], 2)) + '@' +
                str(round(dfks['group'], 2)), horizontalalignment='center', color='b')

        badrate_avg = sum(dfkslift.bad)/sum(dfkslift.good+dfkslift.bad)
        fig, ax = plt.subplot(2, 1, 2)
        ax.plot(dfkslift.group, dfkslift.cumbadrate, 'k-')
        ax.plot([0, 1], [badrate_avg, badrate_avg], 'r--')
        fig.gca().set(title='Lift', xlabel='worst % of population', ylabel='group bad rate',
                      xlim=[0, 1], ylim=[0, 1], aspect='equal')
        ax.text(0.7, np.mean(dfkslift.cumbadrate), 'cumulate badrate', horizontalalignment='center')
        ax.text(0.7, badrate_avg, 'mean: ' + str(round(badrate_avg, 2)), horizontalalignment='center')
    return dfkslift


def cal_psi_bin(base_bin_total, expected_bin_total):
    """
    :param base_bin_total: pd.Series, number of values in each bin
    :param expected_bin_total: pd.Series, number of values in each bin
    """
    assert base_bin_total.shape[0] == expected_bin_total.shape[0]
    base_total = base_bin_total.sum()
    expected_total = expected_bin_total.sum()
    psi_list = []
    psi = 0
    for i in range(base_bin_total.shape[0]):
        base = base_bin_total.iloc[i] / base_total
        expected = expected_bin_total.iloc[i] / expected_total
        p = (expected - base) * np.log(expected / (base + 1e-6) + 1e-6)
        psi_list.append(p)
        psi += p
    return psi, psi_list


def check_trend(var_series, time_series, interval, continuous=True):
    """
    :param var_series: pd.Series, 变量
    :param time_series: pd.Series, 时间
    :param interval: int,间隔,以天为单位
    :param continuous: bool,是否是连续性变量
    """
    time_series = pd.to_datetime(time_series)
    first = time_series.min()
    last = time_series.max()
    date_list = []

    def agg_func(input_series):
        num = input_series.shape[0]
        return [input_series.quantile(0.5),
                input_series.mean(),
                input_series.isnull().sum() / num]

    df_values = []
    for i in range(0, (last - first).days - interval, interval):
        left = first + datetime.timedelta(days=i)
        right = left + datetime.timedelta(days=interval)
        date_list.append('[' + left.strftime("%Y-%m-%d") + ' - '
                         + right.strftime("%Y-%m-%d") + ')')
        data_set = var_series[(time_series >= left) & (time_series < right)]
        if continuous:
            df_values.append(agg_func(data_set))
        else:
            df_values.append([data_set.isnull().sum() / data_set.shape[0]])
    left = right
    right = last
    date_list.append('[' + left.strftime("%Y-%m-%d") + ' - '
                     + right.strftime("%Y-%m-%d") + ']')
    data_set = var_series[(time_series >= left) & (time_series <= right)]
    if continuous:
        df_values.append(agg_func(data_set))
        output_df = pd.DataFrame(df_values, index=date_list,
                                 columns=['0.5q', 'mean', 'null_rate'])
    else:
        df_values.append([data_set.isnull().sum() / data_set.shape[0]])
        output_df = pd.DataFrame(df_values, index=date_list,
                                 columns=['null_rate'])
    return output_df


def plot_trend(df_list, val_column, name_list=[''], save_dir=None):
    """
    看不同数据集上，某个变量随时间变化的趋势是否一致
    :param df_list: list of pd.dataframe, dataframe的index为时间
    :param val_column: string, 当name_list==1的是时候是要看趋势的变量名
                               否则为统计量的名称
    :param name_list: list of string, df_list每个元素的名字
    :param save_dir: string, 当前路径下的相对路径
    """
    assert len(df_list) == len(name_list)
    if len(name_list) == 1:
        trend_df = df_list[0]
    else:
        trend_df = df_list[0][[val_column]].rename(
            columns={val_column: name_list[0]})
        for df, name in zip(df_list[1:], name_list[1:]):
            trend_df = trend_df.join(df[val_column].rename(name), how='outer')
    trend_df.plot(marker='o', markersize=3, title=val_column)
    plt.legend()
    # 绘制x轴（阈值）
    effective_indices_num = trend_df.index.shape[0]
    if effective_indices_num > 5:
        # 向下取整
        increment = int(effective_indices_num / 5)
    else:
        increment = 1
    indices = range(0, effective_indices_num, increment)
    plt.xticks(indices, [i for i in trend_df.index[indices]],
               rotation=-20, horizontalalignment='left')
    if save_dir is not None:
        current_dir = os.getcwd()
        new_dir = os.path.join(current_dir, save_dir)
        if not os.path.exists(new_dir):
            os.mkdir(new_dir)
        save_path = os.path.join(new_dir, val_column + '.jpg')
        plt.savefig(save_path, bbox_inches='tight')
        plt.clf()
        plt.close('all')
    else:
        plt.show()


def check_statistics(df, val_list, continuous=True):
    """
    :param df: pd.dataframe
    :param val_list: list of string, 变量名列表
    :param continuous: bool,是否是连续性变量
    """
    def agg_func_co(input_series):
        num = input_series.shape[0]
        result_arr = np.concatenate(
            (input_series.quantile([0.25, 0.5, 0.75]),
             [input_series.max(), input_series.min(),
              input_series.mean(), input_series.isnull().sum() / num]))
        return pd.Series(result_arr,
                         index=['0.25q', '0.5q', '0.75q',
                                'max', 'min', 'mean', 'null_rate'])

    if continuous:
        output_df = df.loc[:, val_list].apply(agg_func_co, axis=0)
    else:
        output_df = df.loc[:, val_list].apply(
            lambda x: pd.Series(x.isnull().sum() / x.shape[0],
                                index=['null_rate']), axis=0)
    return output_df


def plot_statistics(df_list, val_column, statistics_list,
                    plot_type='bar', name_list=[''], save_dir=None):
    """
    看不同数据集上，某个变量的各项统计指标是否一致
    :param df_list: list of pd.dataframe, dataframe的index是统计指标
    :param val_column: string,要查看的变量名
    :param statistics_list: list of string,统计指标列表，必须和dataframe对应
    :param plot_type: string，如果是bar的话为柱状图，否则为折线图
    :param name_list: list of string, df_list里面的每个元素的名字
    :param save_dir: string,当前路径下的相对路径
    """
    assert len(df_list) == len(name_list)
    if len(name_list) == 1:
        df_bar = df_list[0].loc[statistics_list, [val_column]]
    else:
        df_bar = df_list[0].loc[statistics_list, [val_column]].rename(
            columns={val_column: name_list[0]})
        for df, name in zip(df_list[1:], name_list[1:]):
            df_bar = df_bar.join(
                df.loc[statistics_list, val_column].rename(name), how='outer')
    if plot_type == 'bar':
        df_bar.plot(kind='bar', title=val_column)
    else:
        df_bar.plot(marker='o', markersize=3, title=val_column)
    plt.legend()
    if save_dir is not None:
        current_dir = os.getcwd()
        new_dir = os.path.join(current_dir, save_dir)
        if not os.path.exists(new_dir):
            os.mkdir(new_dir)
        save_path = os.path.join(new_dir, val_column + '.jpg')
        plt.savefig(save_path)
        plt.clf()
        plt.close('all')
    else:
        plt.show()


def plot_bin_df(bin_df_list, name_list=[''], save_dir=None):
    """
    :param bin_df_list: list of dataframe
    :param name_list: bin_df_list里面每个元素的名字
    :param save_dir: string, 前路径下的相对路径
    """
    assert len(bin_df_list) == len(name_list)
    if len(name_list) == 1:
        bar_width = 0.9
    else:
        bar_width = 0.8
    cm_light = mpl.colors.ListedColormap(['#A0FFA0', '#FFA0A0', '#A0A0FF'])
    cm_dark = mpl.colors.ListedColormap(['g', 'r', 'b'])
    df_bin_total = pd.DataFrame()
    df_bad_rate = pd.DataFrame()
    for bin_df, name in zip(bin_df_list, name_list):
        bin_pct_series = bin_df['bin_total_rate'].rename(name + ' bin_pct')
        df_bin_total = pd.concat([df_bin_total, bin_pct_series], axis=1)
        if 'bin_bad_rate' in bin_df.columns.tolist():
            bad_rate_series = bin_df['bin_bad_rate'].rename(name + ' bad_rate')
            df_bad_rate = pd.concat([df_bad_rate, bad_rate_series], axis=1)
    df_bin_total.index = bin_df['bin_no']
    title = bin_df['feature_name'][0]
    ax = df_bin_total.plot(kind='bar', cmap=cm_light,
                           width=bar_width, title=title)
    if 'bin_bad_rate' in bin_df.columns.tolist():
        df_bad_rate.plot(ax=ax, marker='o', markersize=5, cmap=cm_dark)
    ax.set_xlim([-0.5, bin_df.shape[0] - 0.5])
    plt.legend()
    plt.xlabel('bin no.')
    if save_dir is not None:
        current_dir = os.getcwd()
        new_dir = os.path.join(current_dir, save_dir)
        if not os.path.exists(new_dir):
            os.mkdir(new_dir)
        save_path = os.path.join(new_dir, title + '.jpg')
        plt.savefig(save_path)
        plt.clf()
        plt.close('all')
    else:
        plt.show()


# def cal_psi_score(base, expectation):
#     """
#     :param base: pd.Series
#     :param expectation: pd.Series
#     """
#     base_column = pd.qcut(base, 10, duplicates='drop')
#     base_bin = pd.DataFrame(base_column.cat.categories, columns=['bin_range'])
#     base_bin = map_to_inf(base_bin, 'bin_range')
#     base_column = base.astype('category').cat.set_categories(base_bin['bin_range'], ordered=True)
#     base_bin_total = base_column.value_counts().sort_index()
#     expectation_column = pd.cut(expectation, base_bin['bin_range'].astype('category').cat.categories)
#     expected_bin_total = expectation_column.value_counts().sort_index()
#     psi, psi_list = cal_psi_bin(base_bin_total, expected_bin_total)
#     return psi, psi_list


def cal_psi_score(actual_array, expected_array,
                  bins=10, quantile=True, detail=False):
    """
    :param actual_array: np.array
    :param expected_array: np.array
    :param bins: int, number_of_bins you want for calculating psi
    :param quantile: bool
    :param detail: bool, if True, print the process of calculation
    """
    # 异常处理，所有取值都相同时, 说明该变量是常量, 返回None
    if np.min(expected_array) == np.max(expected_array):
        return None
    expected_array = pd.Series(expected_array).dropna()
    actual_array = pd.Series(actual_array).dropna()

    """step1: 确定分箱间隔"""
    def scale_range(input_array, scaled_min, scaled_max):
        """
        功能: 对input_array线性放缩至[scaled_min, scaled_max]
        :param input_array: numpy array of original values, 需放缩的原始数列
        :param scaled_min: float, 放缩后的最小值
        :param scaled_max: float, 放缩后的最大值
        :return input_array: numpy array of original values, 放缩后的数列
        """
        input_array += -np.min(input_array) # 此时最小值放缩到0
        if scaled_max == scaled_min:
            raise Exception('放缩后的数列scaled_min = scaled_min, 值为{}, '
                            '请检查expected_array数值！'.format(scaled_max))
        scaled_slope = np.max(input_array) * 1.0 / (scaled_max - scaled_min)
        input_array /= scaled_slope
        input_array += scaled_min
        return input_array

    breakpoints = np.arange(0, bins + 1) / bins * 100  # 等距分箱百分比
    if not quantile:
        # 等距分箱
        breakpoints = scale_range(breakpoints,
                                  np.min(expected_array),
                                  np.max(expected_array))
    else:
        # 等频分箱
        breakpoints = np.stack([np.percentile(expected_array, b)
                                for b in breakpoints])

    """step2: 统计区间内样本占比"""
    def generate_counts(arr, breakpoints):
        """
        功能: Generates counts for each bucket by using the bucket values
        :param arr: ndarray of actual values
        :param breakpoints: list of bucket values
        :return cnt_array: counts for elements in each bucket,
                           length of breakpoints array minus one
        :return score_range_array: 分箱区间
        """
        def count_in_range(input_arr, low, high, start):
            """
            功能: 统计给定区间内的样本数(Counts elements in array between
                 low and high values)
            :param input_arr: ndarray of actual values
            :param low: float, 左边界
            :param high: float, 右边界
            :param start: bool, 取值为Ture时，区间闭合方式[low, high],否则为(low, high]
            :return cnt_in_range: int, 给定区间内的样本数
            """
            if start:
                cnt_in_range = len(np.where(np.logical_and(input_arr >= low,
                                                           input_arr <= high))[0])
            else:
                cnt_in_range = len(np.where(np.logical_and(input_arr > low,
                                                           input_arr <= high))[0])
            return cnt_in_range
        cnt_array = np.zeros(len(breakpoints) - 1)
        range_array = [''] * (len(breakpoints) - 1)
        for i in range(1, len(breakpoints)):
            cnt_array[i - 1] = count_in_range(arr,
                                              breakpoints[i - 1],
                                              breakpoints[i], i == 1)
            if 1 == i:
                range_array[i - 1] = '[' + \
                                     str(round(breakpoints[i - 1], 4)) \
                                     + ',' + str(round(breakpoints[i], 4)) \
                                     + ']'
            else:
                range_array[i - 1] = '(' + \
                                     str(round(breakpoints[i - 1], 4)) \
                                     + ',' + str(round(breakpoints[i], 4)) \
                                     + ']'

        return cnt_array, range_array

    expected_cnt, score_range_array = generate_counts(expected_array,
                                                      breakpoints)[0]
    expected_percents = expected_cnt / len(expected_array)
    actual_cnt = generate_counts(actual_array, breakpoints)[0]
    actual_percents = actual_cnt / len(actual_array)
    delta_percents = actual_percents - expected_percents
    score_range_array = generate_counts(expected_array, breakpoints)[1]

    """step3: 得到最终稳定性指标"""
    def sub_psi(e_perc, a_perc):
        """
        功能: 计算单个分箱内的psi值
        :param e_perc: float, 期望占比
        :param a_perc: float, 实际占比
        :return value: float, 单个分箱内的psi值
        """
        if a_perc == 0: # 实际占比
            a_perc = 0.001
        if e_perc == 0: # 期望占比
            e_perc = 0.001
        value = (e_perc - a_perc) * np.log(e_perc * 1.0 / a_perc)
        return value
    sub_psi_array = [sub_psi(expected_percents[i], actual_percents[i])
                     for i in range(0, len(expected_percents))]
    if detail:
        psi_value = pd.DataFrame()
        psi_value['score_range'] = score_range_array
        psi_value['expecteds'] = expected_cnt
        psi_value['expected(%)'] = expected_percents * 100
        psi_value['actucals'] = actual_cnt
        psi_value['actucal(%)'] = actual_percents * 100
        psi_value['ac - ex(%)'] = delta_percents * 100
        psi_value['actucal(%)'] = psi_value['actucal(%)'].apply(
            lambda x: round(x, 2))
        psi_value['ac - ex(%)'] = psi_value['ac - ex(%)'].apply(
            lambda x: round(x, 2))
        psi_value['ln(ac/ex)'] = psi_value.apply(
            lambda row: np.log((row['actucal(%)'] + 0.001)
                               / (row['expected(%)'] + 0.001)), axis=1)
        psi_value['psi'] = sub_psi_array
        flag = lambda x: '<<<<<<<' if x == psi_value.psi.max() else ''
        psi_value['max'] = psi_value.psi.apply(flag)
        psi_value = psi_value.append([{'score_range': '>>> summary',
                                       'expecteds': sum(expected_cnt),
                                       'expected(%)': 100,
                                       'actucals': sum(actual_cnt),
                                       'actucal(%)': 100,
                                       'ac - ex(%)': np.nan,
                                       'ln(ac/ex)': np.nan,
                                       'psi': np.sum(sub_psi_array),
                                       'max': '<<< result'}], ignore_index=True)
    else:
        psi_value = np.sum(sub_psi_array)
    return psi_value


def check_monotone(regroup, col, ascending):
    """
    :param regroup: dataframe, bin_df with bin_woe
    :param col: str, the column to be chekcked
    :param ascending: True for ascending, False for descending, None for any one
    """
    if regroup.shape[0] <= 2 and (ascending is None):
        return True
    my_list = regroup[col].tolist()

    # ascending
    flag1 = all(x < y for x, y in zip(my_list, my_list[1:]))
    # descending
    flag2 = all(x > y for x, y in zip(my_list, my_list[1:]))
    if ascending is not None:
        if ascending:
            return flag1
        else:
            return flag2
    else:
        return flag1 or flag2


def chi2_merge(regroup, col, max_bin_num=5, min_binpct=0.0, monoticity=True):
    """
    :param regroup: dataframe [col, bin_total, bin_bad, bin_good, bin_bad_rate]
       rows of col are all list
       indices must be increasing from 0 with increment 1
    :param col: string, col name
    :param max_bin_num: int
    :param min_binpct: float
    :param monoticity: bool. If false, bad_rate could be of no order
    :return: dataframe [col, bin_total, bin_bad, bin_good, bin_bad_rate]
      indices is increasing from 0 with increment 1
    """
    monotone = False
    while not monotone:
        # 按卡方合并箱体
        # 当group_interval的长度大于max_bin时，执行while循环
        while regroup.shape[0] > max_bin_num:
            chi_list = []
            for i in range(regroup.shape[0] - 1):
                # 计算每一对相邻区间的卡方值
                chi_df = regroup.loc[i:i+1, :]
                chi_value = cal_chi2(chi_df, 'bin_bad', 'bin_good', 'bin_total')
                chi_list.append(chi_value)
            # 最小的卡方值的索引,如最小值不止一个则index函数保证取第一个
            best_combined_index = chi_list.index(min(chi_list))
            # 将卡方值最小的一对区间进行合并
            regroup = combine_bin_df(regroup, col, best_combined_index)

        bin_df = regroup.copy()
        # 检查是否有箱只有好样本或者只有坏样本
        min_bad_rate = bin_df['bin_bad_rate'].min()
        max_bad_rate = bin_df['bin_bad_rate'].max()
        while min_bad_rate == 0.0 or max_bad_rate == 1.0:
            # 违约率为1或0的箱体的index. index总是返回列表，因此用[0]取第一个元素
            bad_index = bin_df[bin_df['bin_bad_rate'].isin([0.0, 1.0])].index[0]
            if bad_index == (bin_df.shape[0] - 1):
                bin_df = combine_bin_df(bin_df, col, bad_index-1)
            elif bad_index == 0:
                bin_df = combine_bin_df(bin_df, col, 0)
            else:
                # 计算bad_index和前面一个的箱体的卡方值
                chi_df = bin_df.loc[bad_index-1:bad_index, :]
                chi1 = cal_chi2(chi_df, 'bin_bad', 'bin_good', 'bin_total')
                # 计算bad_index和后一个箱体之间的卡方值
                chi_df = bin_df.loc[bad_index:bad_index+1, :]
                chi2 = cal_chi2(chi_df, 'bin_bad', 'bin_good', 'bin_total')
                if chi1 < chi2:
                    # 当chi1<chi2时, 合并bad_index之前与bad_index对应的箱体
                    bin_df = combine_bin_df(bin_df, col, bad_index-1)
                else:
                    # 当chi1>=chi2时,合并bad_index和bad_index之后对应的箱体
                    bin_df = combine_bin_df(bin_df, col, bad_index)

            # 计算每个区间的违约率, 其index一定是从0开始从小到大编号
            min_bad_rate = bin_df['bin_bad_rate'].min()
            max_bad_rate = bin_df['bin_bad_rate'].max()

        # 检查分箱后的最小占比
        if min_binpct > 0.0:
            # 得出最小的区间占比
            min_pct = (bin_df['bin_total'] / bin_df['bin_total'].sum()).min()
            # 当最小的区间占比小于min_binpct且箱体的个数大于3
            while min_pct < min_binpct and bin_df.shape[0] > 3:
                # 下面的逻辑基本与“检验是否有箱体只有好/坏样本”的一致
                bad_index = \
                    bin_df[(bin_df['bin_total']
                            / bin_df['bin_total'].sum()) == min_pct].index[0]
                if bad_index == (bin_df.shape[0] - 1):
                    bin_df = combine_bin_df(bin_df, col, bad_index-1)
                elif bad_index == 0:
                    bin_df = combine_bin_df(bin_df, col, 0)
                else:
                    chi_df = bin_df.loc[bad_index-1:bad_index, :]
                    chi1 = cal_chi2(chi_df, 'bin_bad', 'bin_good', 'bin_total')
                    chi_df = bin_df.loc[bad_index:bad_index+1, :]
                    chi2 = cal_chi2(chi_df, 'bin_bad', 'bin_good', 'bin_total')
                    if chi1 < chi2:
                        bin_df = combine_bin_df(bin_df, col, bad_index-1)
                    else:
                        bin_df = combine_bin_df(bin_df, col, bad_index)
                min_pct = (bin_df['bin_total'] / bin_df['bin_total'].sum()).min()
        if monoticity:
            bin_df2 = compute_woe_iv(bin_df, 'bin_good', 'bin_bad', 'bin_total')
            monotone = check_monotone(bin_df2, 'bin_woe', None)
            if not monotone:
                print('another chi2 merge is going to be made as the bad rate is not monotonic')
            max_bin_num -= 1
        else:
            break
    return bin_df


def check_bin(regroup, col, continuous_flag, min_binpct=0.05, monoticity=True):
    """
    :param regroup: dataframe [col, bin_total, bin_bad, bin_good, bin_bad_rate]
       indices must be increasing from 0 with increment 1
    :param col: string, col name
    :param min_binpct: float
    :param monoticity: bool. If false, bad_rate could be of no order
    :return: dataframe [col, bin_total, bin_bad, bin_good, bin_bad_rate]
      indices is increasing from 0 with increment 1
    """
    bin_df = regroup.copy()
    # 检查是否有箱只有好样本或者只有坏样本
    min_bad_rate = bin_df['bin_bad_rate'].min()
    max_bad_rate = bin_df['bin_bad_rate'].max()
    while (min_bad_rate == 0.0 or max_bad_rate == 1.0) and bin_df.shape[0] > 2:
        # 违约率为1或0的箱体的index. index总是返回列表，因此用[0]取第一个元素
        bad_index = bin_df[bin_df['bin_bad_rate'].isin([0.0, 1.0])].index[0]
        if bad_index == (bin_df.shape[0] - 1):
            bin_df = combine_bin_df(bin_df, col, bad_index - 1, continuous_flag)
        elif bad_index == 0:
            bin_df = combine_bin_df(bin_df, col, 0, continuous_flag)
        else:
            # 计算bad_index和前面一个的箱体的卡方值
            chi_df = bin_df.loc[bad_index - 1:bad_index, :]
            chi1 = cal_chi2(chi_df, 'bin_bad', 'bin_good', 'bin_total')
            # 计算bad_index和后一个箱体之间的卡方值
            chi_df = bin_df.loc[bad_index:bad_index + 1, :]
            chi2 = cal_chi2(chi_df, 'bin_bad', 'bin_good', 'bin_total')
            if chi1 < chi2:
                # 当chi1<chi2时, 合并bad_index之前与bad_index对应的箱体
                bin_df = combine_bin_df(bin_df, col, bad_index - 1, continuous_flag)
            else:
                # 当chi1>=chi2时,合并bad_index和bad_index之后对应的箱体
                bin_df = combine_bin_df(bin_df, col, bad_index, continuous_flag)

        # 计算每个区间的违约率, 其index一定是从0开始从小到大编号
        min_bad_rate = bin_df['bin_bad_rate'].min()
        max_bad_rate = bin_df['bin_bad_rate'].max()
    if (min_bad_rate == 0.0 or max_bad_rate == 1.0) and bin_df.shape[0] <= 2:
        print('all good or bad, failure')
        return None

    # 检查分箱后的最小占比
    if min_binpct > 0.0:
        # 得出最小的区间占比
        min_pct = (bin_df['bin_total'] / bin_df['bin_total'].sum()).min()
        # 当最小的区间占比小于min_binpct且箱体的个数大于2
        while min_pct < min_binpct and bin_df.shape[0] > 2:
            # 下面的逻辑基本与“检验是否有箱体只有好/坏样本”的一致
            bad_index = \
                bin_df[(bin_df['bin_total']
                        / bin_df['bin_total'].sum()) == min_pct].index[0]
            if bad_index == (bin_df.shape[0] - 1):
                bin_df = combine_bin_df(bin_df, col, bad_index - 1, continuous_flag)
            elif bad_index == 0:
                bin_df = combine_bin_df(bin_df, col, 0, continuous_flag)
            else:
                chi_df = bin_df.loc[bad_index - 1:bad_index, :]
                chi1 = cal_chi2(chi_df, 'bin_bad', 'bin_good', 'bin_total')
                chi_df = bin_df.loc[bad_index:bad_index + 1, :]
                chi2 = cal_chi2(chi_df, 'bin_bad', 'bin_good', 'bin_total')
                if chi1 < chi2:
                    bin_df = combine_bin_df(bin_df, col, bad_index - 1, continuous_flag)
                else:
                    bin_df = combine_bin_df(bin_df, col, bad_index, continuous_flag)
            min_pct = (bin_df['bin_total'] / bin_df['bin_total'].sum()).min()
        if min_pct < min_binpct and bin_df.shape[0] <= 2:
            print('few data in a bin, failure')
            return None

    if monoticity:
        bin_df2 = compute_woe_iv(bin_df, 'bin_good', 'bin_bad', 'bin_total')
        monotone = check_monotone(bin_df2, 'bin_woe', None)
        if not monotone:
            print('woe not monotone, failure')
            return None
    return bin_df


def select_vals_using_iv(dict_of_bin_df, threshold_iv=0.02):
    indices_list = []
    iv_list = []
    for val_name, bin_df in dict_of_bin_df.items():
        indices_list.append(val_name)
        iv_list.append(bin_df['iv'].iloc[0])
    ax = plt.subplot(111)
    iv_series = pd.Series(iv_list, index=indices_list).sort_values(ascending=False)
    selected_vals_series = iv_series[iv_series >= threshold_iv]
    iv_series.plot(kind='bar', ax=ax, color='g', alpha=0.3)
    plt.ylabel('iv')
    plt.show()
    return selected_vals_series


def sub_chisq(ks_points, start, end, bc_gap, bad_rate):
    if start >= end - 1 or sum(ks_points[end][1:]) - sum(ks_points[start][1:]) < 2 * bc_gap:
        return []
    temp = []
    start_good, start_bad = ks_points[start][1], ks_points[start][2]
    good, bad = max(ks_points[end][1] - start_good, 1e-6), max(ks_points[end][2] - start_bad, 1e-6)
    if good == 0 or bad == 0:
        return []
    for j in range(start, end):
        cur_good = ks_points[j][1]
        cur_bad = ks_points[j][2]

        if cur_good + cur_bad > 0:
            left_total = cur_good - start_good + cur_bad - start_bad
            right_total = good - cur_good + bad - cur_bad
            left_bad_expected = int(left_total * bad_rate)
            right_bad_expected = int(right_total * bad_rate)
            left_good_expected = int(left_total * (1 - bad_rate))
            right_good_expected = int(right_total * (1 - bad_rate))
            a11 = 0 if left_good_expected == 0 else np.square(
                (cur_good - start_good) - left_good_expected) / left_good_expected
            a12 = 0 if left_bad_expected == 0 else \
                np.square((cur_bad - start_bad) - left_bad_expected) / left_bad_expected
            a21 = 0 if right_good_expected == 0 else np.square(
                (good - cur_good) - right_good_expected) / right_good_expected
            a22 = 0 if right_bad_expected == 0 else \
                np.square((bad - cur_bad) - right_bad_expected) / right_bad_expected
            chisq = a11 + a12 + a21 + a22

        else:
            chisq = -1
        temp.append((chisq, j - start))
    max_index = max(temp, key=lambda x: x[0])[1]
    while temp[max_index][0] > 0 and \
            (sum(ks_points[max_index + start][1:]) - sum(ks_points[start][1:]) < bc_gap
             or sum(ks_points[end][1:]) - sum(ks_points[max_index + start][1:]) < bc_gap):
        temp[max_index] = (-1, -1)
        ks_points[max_index + start] = (ks_points[max_index + start][0], 0, 0)
        max_index = max(temp, key=lambda x: x[0])[1]

    if temp[max_index][0] <= 0:
        return []
    return [ks_points[max_index + start]] + sub_chisq(ks_points, start, max_index + start, bc_gap, bad_rate) \
           + sub_chisq(ks_points, max_index + start, end, bc_gap, bad_rate)


def best_combine(ks_points, good_bad, piece, monotonicity, lower, upper):
    num_cut_point = len(ks_points)
    num_of_cut = min(piece - 1, num_cut_point)
    cut_points = list(itertools.combinations(range(num_cut_point), num_of_cut))
    sol, max_iv = None, 0
    for cut in cut_points:
        old, iv, iv_list, woe_list = [0, 0], 0, [], []
        for c in cut:
            good_pcnt = (ks_points[c][-2] - old[0]) / good_bad[0]
            bad_pcnt = (ks_points[c][-1] - old[1]) / good_bad[1]
            old = (ks_points[c][-2], ks_points[c][-1])
            woe_list.append(np.log(bad_pcnt / (good_pcnt + 1e-6) + 1e-6))
            iv_list.append(woe_list[-1] * (bad_pcnt - good_pcnt))
            iv += iv_list[-1]
        good_pcnt = (good_bad[0] - old[0]) / good_bad[0]
        bad_pcnt = (good_bad[1] - old[1]) / good_bad[1]
        woe_list.append(np.log(bad_pcnt / (good_pcnt + 1e-6) + 1e-6))
        iv_list.append(woe_list[-1] * (bad_pcnt - good_pcnt))
        iv += iv_list[-1]
        if iv > max_iv:
            flag = True
            flag2 = True
            if monotonicity:
                for i in range(1, len(woe_list)):
                    if woe_list[i] < woe_list[i - 1]:
                        flag = False
                        break
                for i in range(1, len(woe_list)):
                    if woe_list[i] >= woe_list[i - 1]:
                        flag2 = False
                        break
                flag = flag or flag2
            if upper < lower:
                for i in range(1, len(iv_list)):
                    if iv_list[i] < lower or iv_list[i] > upper:
                        flag = False
                        break
            if flag:
                sol = cut
                max_iv = iv
    return sol


def calc_all_information(ks_points, good_bad, sol):
    result, old = [], [0, 0]
    for t in range(len(sol)):
        c = sol[t]
        if t == 0:
            bin_name = pd.Interval(left=float('-inf'), right=ks_points[c][0])
        else:
            bin_name = pd.Interval(left=ks_points[sol[t - 1]][0], right=ks_points[c][0])
        gd_cum, bd_cum = ks_points[c][-2], ks_points[c][-1]
        gd, bd = gd_cum - old[0], bd_cum - old[1]
        total = gd + bd
        bad_rate = bd / total if total > 0 else 0.0
        old = (ks_points[c][-2], ks_points[c][-1])
        result.append([bin_name, total, gd, bd, round(bad_rate, 6)])
    bin_name = pd.Interval(left=ks_points[sol[-1]][0], right=float('inf'))
    gd_cum, bd_cum = float(good_bad[0]), float(good_bad[1])
    gd, bd = gd_cum - old[0], bd_cum - old[1]
    total = gd + bd
    bad_rate = bd / total if total > 0 else 1.0
    result.append([bin_name, total, bd, gd, round(bad_rate, 6)])
    return result


def best_chisq_bin(group, bc_piece, bc_good_bad, bc_gap, bc_strict_monotonicity,
                   bc_iv_lower, bc_iv_upper, merge):
    feat_name = group[0][0]
    good = bc_good_bad[0]
    bad = bc_good_bad[1]
    bad_rate = 1.0 * bad / (bad + good)
    group = sorted(group, key=lambda x: x[1])
    counter = [1.0, 0.0] if group[0][-1] == 0 else [0.0, 1.0]

    ks_points = [(feat_name, counter[0], counter[1])]
    for j in range(1, len(group)):
        if group[j][1] != group[j - 1][1]:
            ks_points.append((group[j - 1][1], counter[0], counter[1]))
        counter[0 if group[j][-1] == 0 else 1] += 1
    if group[-1][1] != ks_points[-1][0]:
        ks_points.append((group[-1][1], counter[0], counter[1]))

    if merge:
        ks_points = sub_chisq(ks_points, 0, len(ks_points) - 1, bc_gap, bad_rate)
        ks_points = sorted(ks_points, key=lambda x: x[0])
        sub_piece = bc_piece
    else:
        sub_piece = len(ks_points) - 1
        ks_points = ks_points[1:-1]

    can_be_bin = False
    for tmp_piece in range(sub_piece - 1):
        sol = best_combine(ks_points, bc_good_bad, sub_piece - tmp_piece,
                           bc_strict_monotonicity, bc_iv_lower, bc_iv_upper)
        if sol:
            can_be_bin = True
            result = calc_all_information(ks_points, bc_good_bad, sol)
            break
    if not can_be_bin:
        print("不能对字段{}进行分箱".format(feat_name))
        return None
    return pd.DataFrame(result, columns=['bin_range', 'bin_total', 'bin_bad', 'bin_good', 'bin_bad_rate'])
